"""Startpoint base classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy as np

from ..C import FVAL, GRAD
from ..objective import NegLogParameterPriors, ObjectiveBase

if TYPE_CHECKING:
    import pypesto


class StartpointMethod(ABC):
    """Startpoint generation, in particular for multi-start optimization.

    Abstract base class, specific sampling method needs to be defined in
    sub-classes.
    """

    @abstractmethod
    def __call__(
        self,
        n_starts: int,
        problem: pypesto.problem.Problem,
    ) -> np.ndarray:
        """Generate startpoints.

        Parameters
        ----------
        n_starts: Number of starts.
        problem: Problem specifying e.g. dimensions, bounds, and guesses.

        Returns
        -------
        xs: Startpoints, shape (n_starts, n_par).
        """


class NoStartpoints(StartpointMethod):
    """Dummy class generating nan points. Useful if no startpoints needed."""

    def __call__(
        self,
        n_starts: int,
        problem: pypesto.problem.Problem,
    ) -> np.ndarray:
        """Generate a (n_starts, dim) nan matrix."""
        startpoints = np.full(shape=(n_starts, problem.dim), fill_value=np.nan)
        return startpoints


class CheckedStartpoints(StartpointMethod, ABC):
    """Startpoints checked for function value and/or gradient finiteness."""

    def __init__(
        self,
        use_guesses: bool = True,
        check_fval: bool = False,
        check_grad: bool = False,
    ):
        """Initialize.

        Parameters
        ----------
        use_guesses:
            Whether to use guesses provided in the problem.
        check_fval:
            Whether to check function values at the startpoint, and resample
            if not finite.
        check_grad:
            Whether to check gradients at the startpoint, and resample
            if not finite.
        """
        self.use_guesses: bool = use_guesses
        self.check_fval: bool = check_fval
        self.check_grad: bool = check_grad

    def __call__(
        self,
        n_starts: int,
        problem: pypesto.problem.Problem,
    ) -> np.ndarray:
        """Generate checked startpoints."""
        # shape: (n_guesses, dim)
        x_guesses = problem.x_guesses
        if not self.use_guesses:
            x_guesses = np.zeros(shape=(0, problem.dim))
        dim = problem.dim
        lb, ub = problem.lb_init, problem.ub_init

        # number of required startpoints
        n_guesses = x_guesses.shape[0]
        n_required = n_starts - n_guesses

        if n_required <= 0:
            return x_guesses[:n_starts, :]

        # apply startpoint method
        x_sampled = self.sample(
            n_starts=n_required, lb=lb, ub=ub, priors=problem.x_priors
        )

        # assemble
        xs = np.zeros(shape=(n_starts, dim))
        xs[0:n_guesses, :] = x_guesses
        xs[n_guesses:n_starts, :] = x_sampled

        # check, resample and order startpoints
        xs = self.check_and_resample(
            xs=xs, lb=lb, ub=ub, objective=problem.objective
        )

        return xs

    @abstractmethod
    def sample(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
        priors: NegLogParameterPriors | None = None,
    ) -> np.ndarray:
        """Actually sample startpoints.

        While in this implementation, `__call__` handles the checking of
        guesses and resampling, this method defines the actual sampling.

        Parameters
        ----------
        n_starts: Number of startpoints to generate.
        lb: Lower parameter bound.
        ub: Upper parameter bound.
        priors: Parameter priors, if available. Some sampling methods may
            utilize this information.

        Returns
        -------
        xs: Startpoints, shape (n_starts, n_par).
        """

    def check_and_resample(
        self,
        xs: np.ndarray,
        lb: np.ndarray,
        ub: np.ndarray,
        objective: ObjectiveBase,
        priors: NegLogParameterPriors | None = None,
    ) -> np.ndarray:
        """Check sampled points for fval, grad, and potentially resample ones.

        Parameters
        ----------
        xs: Startpoints candidates, shape (n_starts, n_par).
        lb: Lower parameter bound.
        ub: Upper parameter bound.
        objective: Objective function, for evaluation.
        priors: Parameter priors, if available.

        Returns
        -------
        xs:
            Checked and potentially partially resampled startpoints,
            shape (n_starts, n_par).
        """
        if not self.check_fval and not self.check_grad:
            return xs

        if self.check_fval and not self.check_grad:
            sensi_orders = (0,)
        elif not self.check_fval and self.check_grad:
            sensi_orders = (1,)
        else:
            sensi_orders = 0, 1

        # track function values for ordering
        fvals = np.empty(shape=(xs.shape[0],))

        # iterate over all startpoint candidates
        for ix, x in enumerate(xs):
            # evaluate candidate
            objective.initialize()
            ret = objective(x, sensi_orders=sensi_orders, return_dict=True)
            fvals[ix] = ret.get(FVAL, np.nan)

            # loop until all requested sensis are finite
            while True:
                # discontinue if all requested sensis are finite
                if (0 not in sensi_orders or np.isfinite(ret[FVAL])) and (
                    1 not in sensi_orders or np.isfinite(ret[GRAD]).all()
                ):
                    break

                # resample a single point
                x = self.sample(n_starts=1, lb=lb, ub=ub, priors=priors)

                # evaluate candidate
                objective.initialize()
                ret = objective(x, sensi_orders=sensi_orders, return_dict=True)
                fvals[ix] = ret.get(FVAL, np.nan)

            # assign permissible value
            xs[ix] = x

        # sort startpoints by function value
        xs_order = np.argsort(fvals)
        xs = xs[xs_order, :]

        return xs


class FunctionStartpoints(CheckedStartpoints):
    """Define startpoints via callable.

    The callable should take the same arguments as the `__call__` method.
    """

    def __init__(
        self,
        function: Callable,
        use_guesses: bool = True,
        check_fval: bool = False,
        check_grad: bool = False,
    ):
        """Initialize.

        Parameters
        ----------
        function: The callable sampling startpoints.
        use_guesses, check_fval, check_grad: As in CheckedStartpoints.
        """
        super().__init__(
            use_guesses=use_guesses,
            check_fval=check_fval,
            check_grad=check_grad,
        )
        self.function: Callable = function

    def sample(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
        priors: NegLogParameterPriors | None = None,
    ) -> np.ndarray:
        """Call function."""
        return self.function(n_starts=n_starts, lb=lb, ub=ub, priors=priors)


def to_startpoint_method(
    maybe_startpoint_method: StartpointMethod | Callable | bool,
) -> StartpointMethod:
    """Create StartpointMethod instance if possible, otherwise raise.

    Parameters
    ----------
    maybe_startpoint_method:
        A StartpointMethod instance, or a Callable as expected by
        FunctionStartpoints.

    Returns
    -------
    startpoint_method:
        A StartpointMethod instance.

    Raises
    ------
    TypeError if arguments cannot be converted to a StartpointMethod.
    """
    if isinstance(maybe_startpoint_method, StartpointMethod):
        return maybe_startpoint_method
    if isinstance(maybe_startpoint_method, Callable):
        return FunctionStartpoints(maybe_startpoint_method)
    if maybe_startpoint_method is False:
        return NoStartpoints()
    raise TypeError(
        "Could not parse startpoint method of type "
        f"{type(maybe_startpoint_method)} to a StartpointMethod.",
    )



import logging
from collections.abc import Callable, Iterable, Sequence
from typing import Optional

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Colormap
from matplotlib.ticker import MaxNLocator

from pypesto.util import delete_nan_inf

from ..C import (
    COLOR,
    INNER_PARAMETERS,
    LOG10,
    WATERFALL_MAX_VALUE,
    InnerParameterType,
)
from ..result import Result
from .clust_color import assign_colors
from .misc import (
    process_parameter_indices,
    process_result_list,
    process_start_indices,
)
from .reference_points import ReferencePoint, create_references

try:
    from ..hierarchical.base_problem import scale_value
    from ..hierarchical.relative import RelativeInnerProblem
    from ..hierarchical.semiquantitative import SemiquantProblem
except ImportError:
    pass

logger = logging.getLogger(__name__)


def parameters(
    results: Result | Sequence[Result],
    ax: matplotlib.axes.Axes | None = None,
    parameter_indices: str | Sequence[int] = "free_only",
    lb: np.ndarray | list[float] | None = None,
    ub: np.ndarray | list[float] | None = None,
    size: tuple[float, float] | None = None,
    reference: list[ReferencePoint] | None = None,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    legends: str | list[str] | None = None,
    balance_alpha: bool = True,
    start_indices: int | Iterable[int] | None = None,
    scale_to_interval: tuple[float, float] | None = None,
    plot_inner_parameters: bool = True,
    log10_scale_hier_sigma: bool = True,
) -> matplotlib.axes.Axes:
    """
    Plot parameter values.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    ax:
        Axes object to use.
    parameter_indices:
        Specifies which parameters should be plotted. Allowed string values are
        'all' (both fixed and free parameters will be plotted)  and
        'free_only' (only free parameters will be plotted)
    lb, ub:
        If not None, override result.problem.lb, problem.problem.ub.
        Dimension either result.problem.dim or result.problem.dim_full.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    reference:
        List of reference points for optimization results, containing at
        least a function value fval
    colors:
        list of colors recognized by matplotlib, or single color
        If not set, clustering is done and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)
    start_indices:
        list of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    scale_to_interval:
        Tuple of bounds to which to scale all parameter values and bounds, or
        ``None`` to use bounds as determined by ``lb, ub``.
    plot_inner_parameters:
        Flag indicating whether to plot inner parameters (default: True).
    log10_scale_hier_sigma:
        Flag indicating whether to scale inner parameters of type
        ``InnerParameterType.SIGMA`` to log10 (default: True).

    Returns
    -------
    ax:
        The plot axes.
    """
    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    if isinstance(parameter_indices, str):
        if parameter_indices == "all":
            parameter_indices = range(0, results[0].problem.dim_full)
        elif parameter_indices == "free_only":
            parameter_indices = results[0].problem.x_free_indices
        else:
            raise ValueError(
                "Permissible values for parameter_indices are "
                "'all', 'free_only' or a list of indices"
            )

    def scale_parameters(x):
        """Scale `x` from [lb, ub] to interval given by `scale_to_interval`."""
        if scale_to_interval is None or scale_to_interval is False:
            return x

        return scale_to_interval[0] + (x - lb) / (ub - lb) * (
            scale_to_interval[1] - scale_to_interval[0]
        )

    for j, result in enumerate(results):
        # handle results and bounds
        (lb, ub, x_labels, fvals, xs, x_axis_label) = handle_inputs(
            result=result,
            lb=lb,
            ub=ub,
            parameter_indices=parameter_indices,
            start_indices=start_indices,
            plot_inner_parameters=plot_inner_parameters,
            log10_scale_hier_sigma=log10_scale_hier_sigma,
        )

        # parse fvals and parameters
        fvals = np.array(fvals)
        # remove nan or inf values
        xs, fvals = delete_nan_inf(
            fvals=fvals,
            x=xs,
            xdim=len(ub) if ub is not None else 1,
            magnitude_bound=WATERFALL_MAX_VALUE,
        )

        lb, ub, xs = map(scale_parameters, (lb, ub, xs))

        # call lowlevel routine
        ax = parameters_lowlevel(
            xs=xs,
            fvals=fvals,
            lb=lb,
            ub=ub,
            x_labels=x_labels,
            x_axis_label=x_axis_label,
            ax=ax,
            size=size,
            colors=colors[j],
            legend_text=legends[j],
            balance_alpha=balance_alpha,
        )

    # parse and apply plotting options
    ref = create_references(references=reference)

    # plot reference points
    for i_ref in ref:
        # reduce parameter vector in reference point, if necessary
        if len(parameter_indices) < results[0].problem.dim_full:
            x_ref = np.array(
                results[0].problem.get_reduced_vector(
                    i_ref["x"], parameter_indices
                )
            )
        else:
            x_ref = np.array(i_ref["x"])
        x_ref = np.reshape(x_ref, (1, x_ref.size))
        x_ref = scale_parameters(x_ref)

        # plot reference parameters using lowlevel routine
        ax = parameters_lowlevel(
            x_ref,
            [i_ref["fval"]],
            ax=ax,
            colors=i_ref["color"],
            linestyle="--",
            legend_text=i_ref.legend,
            balance_alpha=balance_alpha,
        )

    return ax


def parameter_hist(
    result: Result,
    parameter_name: str,
    bins: int | str = "auto",
    ax: Optional["matplotlib.Axes"] = None,
    size: tuple[float] | None = (18.5, 10.5),
    color: COLOR | None = None,
    start_indices: int | list[int] | None = None,
):
    """
    Plot parameter values as a histogram.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'
    parameter_name:
        The name of the parameter that should be plotted
    bins:
        Specifies bins of the histogram
    ax:
        Axes object to use
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    color:
        Color recognized by matplotlib.
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted

    Returns
    -------
    ax:
        The plot axes.
    """
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    xs = result.optimize_result.x

    # reduce number of displayed results
    if isinstance(start_indices, int):
        xs = xs[:start_indices]
    elif start_indices is not None:
        xs = [xs[ind] for ind in start_indices]

    parameter_index = result.problem.x_names.index(parameter_name)
    parameter_values = [x[parameter_index] for x in xs]

    ax.hist(parameter_values, color=color, bins=bins, label=parameter_name)
    ax.set_xlabel(parameter_name)
    ax.set_ylabel("counts")
    ax.set_title(f"{parameter_name}")

    return ax


def parameters_lowlevel(
    xs: np.ndarray,
    fvals: np.ndarray,
    lb: np.ndarray | list[float] | None = None,
    ub: np.ndarray | list[float] | None = None,
    x_labels: Iterable[str] | None = None,
    x_axis_label: str = "Parameter value",
    ax: matplotlib.axes.Axes | None = None,
    size: tuple[float, float] | None = None,
    colors: Sequence[np.ndarray | COLOR] | None = None,
    linestyle: str = "-",
    legend_text: str | None = None,
    balance_alpha: bool = True,
) -> matplotlib.axes.Axes:
    """
    Plot parameters plot using list of parameters.

    Parameters
    ----------
    xs:
        Including optimized parameters for each start that did not result in an infinite fval.
        Shape: (n_starts_successful, dim).
    fvals:
        Function values. Needed to assign cluster colors.
    lb, ub:
        The lower and upper bounds.
    x_labels:
        Labels to be used for the parameters.
    ax:
        Axes object to use.
    size:
        see parameters
    colors:
        A single color recognized by matplotlib or a list of colors, one for each element in 'fvals'.
    linestyle:
        linestyle argument for parameter plot
    legend_text:
        Label for line plots
    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)

    Returns
    -------
    ax:
        The plot axes.
    """

    if size is None:
        # 0.5 inch height per parameter
        size = (18.5, max(xs.shape[1], 1) / 2)

    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # assign colors
    colors = assign_colors(
        vals=fvals, colors=colors, balance_alpha=balance_alpha
    )

    # parameter indices
    parameters_ind = list(range(1, xs.shape[1] + 1))[::-1]

    # plot parameters
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    for j_x, x in reversed(list(enumerate(xs))):
        if j_x == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None
        ax.plot(
            x,
            parameters_ind,
            linestyle,
            color=colors[j_x],
            marker="o",
            label=tmp_legend,
        )

    ax.set_yticks(parameters_ind)
    if x_labels is not None:
        ax.set_yticklabels(x_labels)

    # draw bounds
    parameters_ind = np.array(parameters_ind).flatten()
    if lb is not None:
        lb = np.array(lb, dtype="float64")
        ax.plot(lb.flatten(), parameters_ind, "k--", marker="+")
    if ub is not None:
        ub = np.array(ub, dtype="float64")
        ax.plot(ub.flatten(), parameters_ind, "k--", marker="+")

    ax.set_xlabel(x_axis_label)
    ax.set_ylabel("Parameter")
    ax.set_title("Estimated parameters")
    if legend_text is not None:
        ax.legend()

    return ax


def handle_inputs(
    result: Result,
    parameter_indices: list[int],
    lb: np.ndarray | list[float] | None = None,
    ub: np.ndarray | list[float] | None = None,
    start_indices: int | Iterable[int] | None = None,
    plot_inner_parameters: bool = False,
    log10_scale_hier_sigma: bool = True,
) -> tuple[np.ndarray, np.ndarray, list[str], np.ndarray, list[np.ndarray]]:
    """
    Compute the correct bounds for the parameter indices to be plotted.

    Outputs the corresponding parameters and their labels.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'.
    parameter_indices:
        Specifies which parameters should be plotted.
    lb, ub:
        If not None, override result.problem.lb, problem.problem.ub.
        Dimension either result.problem.dim or result.problem.dim_full.
    start_indices:
        list of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    plot_inner_parameters:
        Flag indicating whether inner parameters should be plotted.
    log10_scale_hier_sigma:
        Flag indicating whether to scale inner parameters of type
        ``InnerParameterType.SIGMA`` to log10 (default: True).

    Returns
    -------
    lb, ub:
        Dimension either result.problem.dim or result.problem.dim_full.
    x_labels:
        ytick labels to be applied later on
    fvals:
        objective function values which are needed for plotting later
    xs:
        parameter values which will be plotted later
    x_axis_label:
        label for the x-axis
    """
    # retrieve results
    fvals = result.optimize_result.fval
    xs = result.optimize_result.x

    # retrieve inner parameters in case of hierarchical optimization
    (
        inner_xs,
        inner_xs_names,
        inner_xs_scales,
        inner_lb,
        inner_ub,
    ) = _handle_inner_inputs(result, log10_scale_hier_sigma)

    # parse indices which should be plotted
    if start_indices is not None:
        start_indices = process_start_indices(result, start_indices)

        # reduce number of displayed results
        xs_out = [xs[ind] for ind in start_indices]
        fvals_out = [fvals[ind] for ind in start_indices]
        if inner_xs is not None and plot_inner_parameters:
            inner_xs_out = [inner_xs[ind] for ind in start_indices]
    else:
        # use non-reduced versions
        xs_out = xs
        fvals_out = fvals
        if inner_xs is not None and plot_inner_parameters:
            inner_xs_out = inner_xs

    # get bounds
    if lb is None:
        lb = result.problem.lb_full
    if ub is None:
        ub = result.problem.ub_full

    # get labels as x_names and scales
    x_labels = list(
        zip(result.problem.x_names, result.problem.x_scales, strict=True)
    )

    # handle fixed and free indices
    if len(parameter_indices) < result.problem.dim_full:
        for ix, x in enumerate(xs_out):
            xs_out[ix] = result.problem.get_reduced_vector(
                x, parameter_indices
            )
        lb = result.problem.get_reduced_vector(lb, parameter_indices)
        ub = result.problem.get_reduced_vector(ub, parameter_indices)
        x_labels = [x_labels[int(i)] for i in parameter_indices]
    else:
        lb = result.problem.lb_full
        ub = result.problem.ub_full

    if inner_xs is not None and plot_inner_parameters:
        lb = np.concatenate([lb, inner_lb])
        ub = np.concatenate([ub, inner_ub])
        inner_xs_labels = list(
            zip(inner_xs_names, inner_xs_scales, strict=True)
        )
        x_labels = x_labels + inner_xs_labels
        xs_out = [
            np.concatenate([x, inner_x]) if x is not None else None
            for x, inner_x in zip(xs_out, inner_xs_out, strict=True)
        ]

    # If all the scales are the same, put it in the x_axis_label
    if len({x_scale for _, x_scale in x_labels}) == 1:
        x_axis_label = "Parameter value (" + x_labels[0][1] + ")"
        x_labels = [x_name for x_name, _ in x_labels]
    else:
        x_axis_label = "Parameter value"
        x_labels = [f"{x_name} ({x_scale})" for x_name, x_scale in x_labels]

    return lb, ub, x_labels, fvals_out, xs_out, x_axis_label


def _handle_inner_inputs(
    result: Result,
    log10_scale_hier_sigma: bool = True,
) -> (
    tuple[None, None, None, None, None]
    | tuple[list[np.ndarray], list[str], list[str], np.ndarray, np.ndarray]
):
    """Handle inner parameters from hierarchical optimization, if available.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'.
    log10_scale_hier_sigma:
        Flag indicating whether to scale inner parameters of type
        ``InnerParameterType.SIGMA`` to log10 (default: True).

    Returns
    -------
    inner_xs:
        Inner parameter values which will be appended to xs.
    inner_xs_names:
        Inner parameter names.
    inner_xs_scales:
        Inner parameter scales.
    inner_lb:
        Inner parameter lower bounds.
    inner_ub:
        Inner parameter upper bounds.
    """
    inner_xs = [
        res.get(INNER_PARAMETERS, None) for res in result.optimize_result.list
    ]
    inner_xs_names = None
    inner_xs_scales = None
    inner_lb = None
    inner_ub = None

    from ..problem import HierarchicalProblem

    if any(inner_x is not None for inner_x in inner_xs) and isinstance(
        result.problem, HierarchicalProblem
    ):
        inner_xs_names = result.problem.inner_x_names
        # replace None with a list of nans
        inner_xs = [
            (
                np.full(len(inner_xs_names), np.nan)
                if inner_xs_for_start is None
                else np.asarray(inner_xs_for_start)
            )
            for inner_xs_for_start in inner_xs
        ]
        # set bounds for inner parameters
        inner_lb = result.problem.inner_lb
        inner_ub = result.problem.inner_ub

        # Scale inner parameter bounds according to their parameters scales
        inner_xs_scales = result.problem.inner_scales

        if log10_scale_hier_sigma:
            inner_problems_with_sigma = [
                inner_calculator.inner_problem
                for inner_calculator in result.problem.objective.calculator.inner_calculators
                if isinstance(
                    inner_calculator.inner_problem, RelativeInnerProblem
                )
                or isinstance(inner_calculator.inner_problem, SemiquantProblem)
            ]
            for inner_problem in inner_problems_with_sigma:
                for inner_x_idx, inner_x_name in enumerate(inner_xs_names):
                    if (inner_x_name in inner_problem.get_x_ids()) and (
                        inner_problem.get_for_id(
                            inner_x_name
                        ).inner_parameter_type
                        == InnerParameterType.SIGMA
                    ):
                        # Scale all values, lower and upper bounds
                        for inner_x_for_start in inner_xs:
                            inner_x_for_start[inner_x_idx] = scale_value(
                                inner_x_for_start[inner_x_idx], LOG10
                            )
                        inner_xs_scales[inner_x_idx] = LOG10

        for inner_x_idx, inner_scale in enumerate(inner_xs_scales):
            inner_lb[inner_x_idx] = scale_value(
                inner_lb[inner_x_idx], inner_scale
            )
            inner_ub[inner_x_idx] = scale_value(
                inner_ub[inner_x_idx], inner_scale
            )

    if inner_xs_names is None:
        inner_xs = None

    return inner_xs, inner_xs_names, inner_xs_scales, inner_lb, inner_ub


def parameters_correlation_matrix(
    result: Result,
    parameter_indices: str | Sequence[int] = "free_only",
    start_indices: int | Iterable[int] | None = None,
    method: str | Callable = "pearson",
    cluster: bool = True,
    cmap: Colormap | str = "bwr",
    return_table: bool = False,
    heatmap_kwargs: dict | None = None,
    size: tuple[float, float] | None = None,
) -> matplotlib.axes.Axes:
    """
    Plot correlation of optimized parameters.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'
    parameter_indices:
        List of integers specifying the parameters to be considered.
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    method:
        The method to compute correlation. Allowed values are ``pearson``,
        ``kendall``, ``spearman`` or a callable.
    cluster:
        Whether to cluster the correlation matrix.
    cmap:
        Colormap to use for the heatmap. Defaults to 'bwr'.
    return_table:
        Whether to return the parameter table additionally for further
        inspection.
    heatmap_kwargs:
        Additional keyword arguments to :func:`seaborn.heatmap`.
    size:
        Figure size (width, height) in inches.

    Returns
    -------
    ax:
        The plot axis.
    """
    import seaborn as sns

    start_indices = process_start_indices(
        start_indices=start_indices, result=result
    )
    parameter_indices = process_parameter_indices(
        parameter_indices=parameter_indices, result=result
    )
    # put all parameters into a dataframe, where columns are parameters
    parameters = [
        result.optimize_result[i_start]["x"][parameter_indices]
        for i_start in start_indices
    ]
    x_labels = [
        result.problem.x_names[parameter_index]
        for parameter_index in parameter_indices
    ]
    df = pd.DataFrame(parameters, columns=x_labels)
    corr_matrix = df.corr(method=method)
    heatmap_kwargs = {
        "data": corr_matrix,
        "yticklabels": True,
        "vmin": -1,
        "vmax": 1,
        "cmap": cmap,
        "linewidth": 1,
    } | (heatmap_kwargs or {})
    if cluster:
        if size is not None:
            heatmap_kwargs["figsize"] = size
        ax = sns.clustermap(**heatmap_kwargs)
    else:
        ax = sns.heatmap(**heatmap_kwargs)
        if size is not None:
            ax.figure.set_size_inches(*size)
    if return_table:
        return ax, df
    return ax


def optimization_scatter(
    result: Result,
    parameter_indices: str | Sequence[int] = "free_only",
    start_indices: int | Iterable[int] | None = None,
    diag_kind: str = "kde",
    suptitle: str = None,
    size: tuple[float, float] = None,
    show_bounds: bool = False,
):
    """
    Plot a scatter plot of all pairs of parameters for the given starts.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'.
    parameter_indices:
        List of integers specifying the parameters to be considered.
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted.
    diag_kind:
        Visualization mode for marginal densities {‘auto’, ‘hist’, ‘kde’,
        None}.
    suptitle:
        Title of the plot.
    size:
        Size of the plot.
    show_bounds:
        Whether to show the parameter bounds.

    Returns
    -------
    ax:
        The plot axis.
    """
    import seaborn as sns

    start_indices = process_start_indices(
        start_indices=start_indices, result=result
    )
    parameter_indices = process_parameter_indices(
        parameter_indices=parameter_indices, result=result
    )
    # put all parameters into a dataframe, where columns are parameters
    parameters = [
        result.optimize_result[i_start]["x"][parameter_indices]
        for i_start in start_indices
    ]
    x_labels = [
        result.problem.x_names[parameter_index]
        for parameter_index in parameter_indices
    ]
    df = pd.DataFrame(parameters, columns=x_labels)

    sns.set(style="ticks")

    ax = sns.pairplot(
        df,
        diag_kind=diag_kind,
    )

    if size is not None:
        ax.fig.set_size_inches(size)
    if suptitle:
        ax.fig.suptitle(suptitle)
    if show_bounds:
        # set bounds of plot to parameter bounds. Only use diagonal as
        # sns.PairGrid has sharex,sharey = True by default.
        for i_axis, axis in enumerate(np.diag(ax.axes)):
            axis.set_xlim(result.problem.lb[i_axis], result.problem.ub[i_axis])
            axis.set_ylim(result.problem.lb[i_axis], result.problem.ub[i_axis])

    return ax



"""RoadRunner calculator for PEtab problems.

Handles all RoadRunner.simulate calls, calculates likelihoods and residuals.
"""

from __future__ import annotations

import numbers
from collections.abc import Sequence

import numpy as np

from ...C import (
    FVAL,
    MODE_FUN,
    MODE_RES,
    RES,
    ROADRUNNER_LLH,
    ROADRUNNER_SIMULATION,
    TIME,
    ModeType,
)
from .utils import (
    ExpData,
    SolverOptions,
    simulation_to_measurement_df,
    unscale_parameters,
)

try:
    import petab.v1 as petab
    from petab.v1.parameter_mapping import ParMappingDictQuadruple
except ImportError:
    petab = None
try:
    import roadrunner
except ImportError:
    roadrunner = None

LLH_TYPES = {
    "lin_normal": lambda measurement, simulation, sigma: (
        -0.5
        * (
            np.log(2 * np.pi * (sigma**2))
            + ((measurement - simulation) / sigma) ** 2
        )
    ),
    "log_normal": lambda measurement, simulation, sigma: (
        -0.5
        * (
            np.log(2 * np.pi * (sigma**2) * (measurement**2))
            + ((np.log(measurement) - np.log(simulation)) / sigma) ** 2
        )
    ),
    "log10_normal": lambda measurement, simulation, sigma: (
        -0.5
        * (
            np.log(2 * np.pi * (sigma**2) * (measurement**2) * np.log(10) ** 2)
            + ((np.log10(measurement) - np.log10(simulation)) / sigma) ** 2
        )
    ),
    "lin_laplace": lambda measurement, simulation, sigma: (
        -np.log(2 * sigma) - (np.abs(measurement - simulation) / sigma)
    ),
    "log_laplace": lambda measurement, simulation, sigma: (
        -np.log(2 * sigma * simulation)
        - (np.abs(np.log(measurement) - np.log(simulation)) / sigma)
    ),
    "log10_laplace": lambda measurement, simulation, sigma: (
        -np.log(2 * sigma * simulation * np.log(10))
        - (np.abs(np.log10(measurement) - np.log10(simulation)) / sigma)
    ),
}


class RoadRunnerCalculator:
    """Class to handle RoadRunner simulation and obtain objective value."""

    def __call__(
        self,
        x_dct: dict,  # TODO: sensi_order support
        mode: ModeType,
        roadrunner_instance: roadrunner.RoadRunner,
        edatas: list[ExpData],
        x_ids: Sequence[str],
        parameter_mapping: list[ParMappingDictQuadruple],
        petab_problem: petab.Problem,
        solver_options: SolverOptions | None = None,
    ):
        """Perform the RoadRunner call and obtain objective function values.

        Parameters
        ----------
        x_dct:
            Parameter dictionary.
        mode:
            Mode of the call.
        roadrunner_instance:
            RoadRunner instance.
        edatas:
            List of ExpData.
        x_ids:
            Sequence of parameter IDs.
        parameter_mapping:
            Parameter parameter_mapping.
        petab_problem:
            PEtab problem.
        solver_options:
            Solver options of the roadrunner instance Integrator. These will
            modify the roadrunner instance inplace.

        Returns
        -------
        Tuple of objective function values.
        """
        # sanity check that edatas and conditions are consistent
        if len(edatas) != len(parameter_mapping):
            raise ValueError(
                "Number of edatas and conditions are not consistent."
            )
        if solver_options is None:
            solver_options = SolverOptions()
        # apply solver options
        solver_options.apply_to_roadrunner(roadrunner_instance)
        simulation_results = {}
        llh_tot = 0
        for edata, mapping_per_condition in zip(
            edatas, parameter_mapping, strict=True
        ):
            sim_res, llh = self.simulate_per_condition(
                x_dct, roadrunner_instance, edata, mapping_per_condition
            )
            simulation_results[edata.condition_id] = sim_res
            llh_tot += llh

        if mode == MODE_FUN:
            return {
                FVAL: -llh_tot,
                ROADRUNNER_SIMULATION: simulation_results,
                ROADRUNNER_LLH: llh_tot,
            }
        if mode == MODE_RES:  # TODO: speed up by not using pandas
            simulation_df = simulation_to_measurement_df(
                simulation_results, petab_problem.measurement_df
            )
            res_df = petab.calculate_residuals(
                petab_problem.measurement_df,
                simulation_df,
                petab_problem.observable_df,
                petab_problem.parameter_df,
            )
            return {
                RES: res_df,
                ROADRUNNER_SIMULATION: simulation_results,
                FVAL: -llh_tot,
            }

    def simulate_per_condition(
        self,
        x_dct: dict,
        roadrunner_instance: roadrunner.RoadRunner,
        edata: ExpData,
        parameter_mapping_per_condition: ParMappingDictQuadruple,
    ) -> tuple[np.ndarray, float]:
        """Simulate the model for a single condition.

        Parameters
        ----------
        x_dct:
            Parameter dictionary.
        roadrunner_instance:
            RoadRunner instance.
        edata:
            ExpData of a single condition.
        parameter_mapping_per_condition:
            Parameter parameter_mapping for a single condition.

        Returns
        -------
        Tuple of simulation results in form of a numpy array and the
        negative log-likelihood.
        """
        # get timepoints and outputs to simulate
        timepoints = list(edata.timepoints)
        if timepoints[0] != 0.0:
            timepoints = [0.0] + timepoints
        if len(timepoints) == 1:
            timepoints = [0.0] + timepoints
        observables_ids = edata.get_observable_ids()
        # steady state stuff
        steady_state_calculations = False
        state_variables = roadrunner_instance.model.getFloatingSpeciesIds()
        # some states might be hidden as parameters with rate rules
        rate_rule_ids = roadrunner_instance.getRateRuleIds()
        state_variables += [
            rate_rule_id
            for rate_rule_id in rate_rule_ids
            if rate_rule_id not in state_variables
        ]
        # obs_ss = []  # TODO: add them to return values with info
        state_ss = []

        # if the first and third parameter mappings are not empty, we need
        # to pre-equlibrate the model
        if (
            parameter_mapping_per_condition[0]
            and parameter_mapping_per_condition[2]
        ):
            steady_state_calculations = True
            roadrunner_instance.conservedMoietyAnalysis = True
            self.fill_in_parameters(
                x_dct,
                roadrunner_instance,
                parameter_mapping_per_condition,
                preeq=True,
            )
            roadrunner_instance.setSteadyStateSolver("newton")
            # allow simulation to reach steady state
            roadrunner_instance.getSteadyStateSolver().setValue(
                "allow_presimulation", True
            )
            roadrunner_instance.getSteadyStateSolver().setValue(
                "presimulation_maximum_steps", 1000
            )
            roadrunner_instance.getSteadyStateSolver().setValue(
                "presimulation_time", 1000
            )
            # steady state output = observables + state variables
            steady_state_selections = observables_ids + state_variables
            roadrunner_instance.steadyStateSelections = steady_state_selections
            steady_state = roadrunner_instance.getSteadyStateValuesNamedArray()
            # we split the steady state into observables and state variables
            # obs_ss = steady_state[:, : len(observables_ids)].flatten()
            state_ss = steady_state[:, len(observables_ids) :].flatten()
            # turn off conserved moiety analysis
            roadrunner_instance.conservedMoietyAnalysis = False
            # reset the model
            roadrunner_instance.reset()
        # set parameters
        par_map = self.fill_in_parameters(
            x_dct, roadrunner_instance, parameter_mapping_per_condition
        )
        # if steady state calculations are required, set state variables
        if steady_state_calculations:
            roadrunner_instance.setValues(state_variables, state_ss)
            # fill in overriden species
            self.fill_in_parameters(
                x_dct,
                roadrunner_instance,
                parameter_mapping_per_condition,
                filling_mode="only_species",
            )

        sim_res = roadrunner_instance.simulate(
            times=timepoints, selections=[TIME] + observables_ids
        )

        llhs = calculate_llh(sim_res, edata, par_map, roadrunner_instance)

        # reset the model
        roadrunner_instance.reset()

        return sim_res, llhs

    def fill_in_parameters(
        self,
        problem_parameters: dict,
        roadrunner_instance: roadrunner.RoadRunner | None = None,
        parameter_mapping: ParMappingDictQuadruple | None = None,
        preeq: bool = False,
        filling_mode: str | None = None,
    ) -> dict:
        """Fill in parameters into the roadrunner instance.

        Parameters
        ----------
        roadrunner_instance:
            RoadRunner instance to fill in parameters
        problem_parameters:
            Problem parameters as parameterId=>value dict. Only
            parameters included here will be set. Remaining parameters will
            be used as already set in `amici_model` and `edata`.
        parameter_mapping:
            Parameter mapping for current condition. Quadruple of dicts,
            where the first dict contains the parameter mapping for pre-
            equilibration, the second dict contains the parameter mapping
            for the simulation, the third and fourth dict contain the scaling
            factors for the pre-equilibration and simulation, respectively.
        preeq:
            Whether to fill in parameters for pre-equilibration.
        filling_mode:
            Which parameters to fill in. If None or "all",
            all parameters are filled in.
            Other options are "only_parameters" and "only_species".

        Returns
        -------
        dict:
            Mapping of parameter IDs to values.
        """
        if filling_mode is None:
            filling_mode = "all"
        # check for valid filling modes
        if filling_mode not in ["all", "only_parameters", "only_species"]:
            raise ValueError(
                "Invalid filling mode. Choose from 'all', "
                "'only_parameters', 'only_species'."
            )
        mapping = parameter_mapping[1]  # default: simulation condition mapping
        scaling = parameter_mapping[3]  # default: simulation condition scaling
        if preeq:
            mapping = parameter_mapping[0]  # pre-equilibration mapping
            scaling = parameter_mapping[2]  # pre-equilibration scaling

        # Parameter parameter_mapping may contain parameter_ids as values,
        # these *must* be replaced

        def _get_par(model_par, val):
            """Get parameter value from problem_parameters and mapping.

            Replace parameter IDs in parameter_mapping dicts by values from
            problem_parameters where necessary
            """
            if isinstance(val, str):
                try:
                    # estimated parameter
                    return problem_parameters[val]
                except KeyError:
                    # condition table overrides must have been handled already,
                    # e.g. by the PEtab parameter parameter_mapping, but
                    # parameters from InitialAssignments may still be present.
                    if mapping[val] == model_par:
                        # prevent infinite recursion
                        raise
                    return _get_par(val, mapping[val])
            if model_par in problem_parameters:
                # user-provided
                return problem_parameters[model_par]
            # prevent nan-propagation in derivative
            if np.isnan(val):
                return 0.0
            # constant value
            return val

        mapping_values = {
            key: _get_par(key, val) for key, val in mapping.items()
        }
        # we assume the parameters to be given in the scale defined in the
        # petab problem. Thus, they need to be unscaled.
        mapping_values = unscale_parameters(mapping_values, scaling)
        # seperate the parameters into ones that overwrite species and others
        mapping_params = {}
        mapping_species = {}
        for key, value in mapping_values.items():
            if key in roadrunner_instance.model.getFloatingSpeciesIds():
                # values that originally were NaN are not set
                if isinstance(mapping[key], str) or not np.isnan(mapping[key]):
                    mapping_species[key] = float(value)
            else:
                mapping_params[key] = value

        if filling_mode == "only_parameters" or filling_mode == "all":
            # set parameters.
            roadrunner_instance.setValues(mapping_params)
            # reset is necessary to apply the changes to initial assignments
            roadrunner_instance.reset()
        if filling_mode == "only_species" or filling_mode == "all":
            # set species
            roadrunner_instance.setValues(mapping_species)
        return mapping_values


def calculate_llh(
    simulations: np.ndarray,
    edata: ExpData,
    parameter_mapping: dict,
    roadrunner_instance: roadrunner.RoadRunner,
) -> float:
    """Calculate the negative log-likelihood for a single condition.

    Parameters
    ----------
    simulations:
        Simulations of condition.
    edata:
        ExpData of a single condition.
    parameter_mapping:
        Parameter mapping for the condition.
    roadrunner_instance:
        RoadRunner instance. Needed to retrieve complex formulae.

    Returns
    -------
    float:
        Negative log-likelihood.
    """
    # if 0 is not in timepoints, remove the first row of the simulation
    if 0.0 not in edata.timepoints:
        simulations = simulations[1:, :]

    if not np.array_equal(simulations[:, 0], edata.timepoints):
        raise ValueError(
            "Simulation and Measurement have different timepoints."
        )
    # check that simulation and condition have same dimensions and timepoints
    if simulations.shape != edata.measurements.shape:
        raise ValueError(
            "Simulation and Measurement have different dimensions."
        )
    # we can now drop the timepoints
    simulations = simulations[:, 1:]
    measurements = edata.measurements[:, 1:]

    def _fill_in_noise_formula(noise_formula):
        """Fill in the noise formula."""
        if isinstance(noise_formula, numbers.Number):
            return float(noise_formula)
        # if it is not a number, it is assumed to be a string
        if noise_formula in parameter_mapping.keys():
            return parameter_mapping[noise_formula]
        # if the string starts with "noiseFormula_" it is saved in the model
        if noise_formula.startswith("noiseFormula_"):
            return roadrunner_instance.getValue(noise_formula)

    # replace noise formula with actual value from mapping
    noise_formulae = np.array(
        [_fill_in_noise_formula(formula) for formula in edata.noise_formulae]
    )
    # check that the rows of noise are the columns of the simulation
    if noise_formulae.shape[0] != simulations.shape[1]:
        raise ValueError("Noise and Simulation have different dimensions.")
    # per observable, decide on the llh function based on the noise dist
    llhs = np.array(
        [
            LLH_TYPES[noise_dist](
                measurements[:, i], simulations[:, i], noise_formulae[i]
            )
            for i, noise_dist in enumerate(edata.noise_distributions)
        ]
    ).transpose()
    # check whether all nan values in llhs coincide with nan measurements
    if not np.all(np.isnan(llhs) == np.isnan(measurements)):
        return np.nan

    # sum over all observables, ignoring nans
    llhs = np.nansum(llhs)

    return llhs



from collections.abc import Sequence
from pathlib import Path

import numpy as np

from ..C import (
    CPU_TIME_TOTAL,
    POSTEQ_CPU_TIME,
    POSTEQ_CPU_TIME_BACKWARD,
    PREEQ_CPU_TIME,
    PREEQ_CPU_TIME_BACKWARD,
    RDATAS,
)
from .csv import CsvHistory
from .hdf5 import Hdf5History
from .options import HistoryOptions
from .util import trace_wrap


class Hdf5AmiciHistory(Hdf5History):
    """
    Stores history extended by AMICI-specific time traces in an HDF5 file.

    Stores AMICI-specific traces of total simulation time, pre-equilibration
    time and post-equilibration time.

    Parameters
    ----------
    id:
        Id of the history
    file:
        HDF5 file name.
    options:
        History options. Defaults to ``None``.
    """

    def __init__(
        self,
        id: str,
        file: str | Path,
        options: HistoryOptions | dict | None = None,
    ):
        super().__init__(id, file, options=options)

    @staticmethod
    def load(
        id: str,
        file: str | Path,
        options: HistoryOptions | dict = None,
    ) -> "Hdf5AmiciHistory":
        """Load the History object from memory."""
        history = Hdf5AmiciHistory(id=id, file=file, options=options)
        if options is None:
            history.recover_options(file)
        return history

    @staticmethod
    def _simulation_to_values(x, result, used_time):
        values = Hdf5History._simulation_to_values(x, result, used_time)
        # default unit for time in amici is [ms], converted to [s]
        values |= {
            key: sum([rdata[key] for rdata in result[RDATAS]]) * 0.001
            for key in (
                CPU_TIME_TOTAL,
                PREEQ_CPU_TIME,
                PREEQ_CPU_TIME_BACKWARD,
                POSTEQ_CPU_TIME,
                POSTEQ_CPU_TIME_BACKWARD,
            )
        }
        return values

    @trace_wrap
    def get_cpu_time_total_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative simulation CPU time [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return self._get_hdf5_entries(CPU_TIME_TOTAL, ix)

    @trace_wrap
    def get_preeq_time_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative pre-equilibration time, [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return self._get_hdf5_entries(PREEQ_CPU_TIME, ix)

    @trace_wrap
    def get_preeq_timeB_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative pre-equilibration time of the backward problem, [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return self._get_hdf5_entries(PREEQ_CPU_TIME_BACKWARD, ix)

    @trace_wrap
    def get_posteq_time_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative post-equilibration time [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return self._get_hdf5_entries(POSTEQ_CPU_TIME, ix)

    @trace_wrap
    def get_posteq_timeB_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative post-equilibration time of the backward problem [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return self._get_hdf5_entries(POSTEQ_CPU_TIME_BACKWARD, ix)


class CsvAmiciHistory(CsvHistory):
    """
    Stores history extended by AMICI-specific time traces in a CSV file.

    Stores AMICI-specific traces of total simulation time, pre-equilibration
    time and post-equilibration time.

    Parameters
    ----------
    file:
        CSV file name.
    x_names:
        Parameter names.
    options:
        History options.
    load_from_file:
        If True, history will be initialized from data in the specified file.
    """

    def __init__(
        self,
        file: str,
        x_names: Sequence[str] = None,
        options: HistoryOptions | dict = None,
        load_from_file: bool = False,
    ):
        super().__init__(file, x_names, options, load_from_file=load_from_file)

    def _trace_columns(self) -> list[tuple]:
        columns = super()._trace_columns()
        return columns + [
            (c, np.nan)
            for c in [
                CPU_TIME_TOTAL,
                PREEQ_CPU_TIME,
                PREEQ_CPU_TIME_BACKWARD,
                POSTEQ_CPU_TIME,
                POSTEQ_CPU_TIME_BACKWARD,
            ]
        ]

    def _simulation_to_values(self, result, used_time):
        values = super()._simulation_to_values(result, used_time)
        # default unit for time in amici is [ms], converted to [s]
        values |= {
            key: sum([rdata[key] for rdata in result[RDATAS]]) * 0.001
            for key in (
                CPU_TIME_TOTAL,
                PREEQ_CPU_TIME,
                PREEQ_CPU_TIME_BACKWARD,
                POSTEQ_CPU_TIME,
                POSTEQ_CPU_TIME_BACKWARD,
            )
        }
        return values

    @trace_wrap
    def get_cpu_time_total_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative simulation CPU time [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return list(self._trace[(CPU_TIME_TOTAL, np.nan)].values[ix])

    @trace_wrap
    def get_preeq_time_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative pre-equilibration time [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return list(self._trace[(PREEQ_CPU_TIME, np.nan)].values[ix])

    @trace_wrap
    def get_preeq_timeB_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative pre-equilibration time of the backward problem [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return list(self._trace[(PREEQ_CPU_TIME_BACKWARD, np.nan)].values[ix])

    @trace_wrap
    def get_posteq_time_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative post-equilibration time [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return list(self._trace[(POSTEQ_CPU_TIME, np.nan)].values[ix])

    @trace_wrap
    def get_posteq_timeB_trace(
        self, ix: int | Sequence[int] | None = None, trim: bool = False
    ) -> Sequence[float] | float:
        """
        Cumulative post-equilibration time of the backward problem [s].

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        return list(self._trace[(POSTEQ_CPU_TIME_BACKWARD, np.nan)].values[ix])



"""Functions for variational inference accessible to the user. Currently only pymc is supported."""

import logging
from collections.abc import Callable
from time import process_time

import numpy as np

from ..problem import Problem
from ..result import Result
from ..sample.util import bound_n_samples_from_env
from ..store import autosave
from .pymc import PymcVariational

logger = logging.getLogger(__name__)


def variational_fit(
    problem: Problem,
    n_iterations: int,
    method: str = "advi",
    n_samples: int | None = None,
    random_seed: int | None = None,
    start_sigma: dict[str, np.ndarray] | None = None,
    x0: np.ndarray | list[np.ndarray] = None,
    result: Result = None,
    filename: str | Callable | None = None,
    overwrite: bool = False,
    **kwargs,
) -> Result:
    """
    Call to do parameter sampling.

    Parameters
    ----------
    problem:
        The problem to be solved. If None is provided, a
        :class:`pypesto.AdaptiveMetropolisSampler` is used.
    n_iterations:
        Number of iterations for the optimization.
    method: str or :class:`Inference` of pymc (only interface currently supported)
        string name is case-insensitive in:
            -   'advi'  for ADVI
            -   'fullrank_advi'  for FullRankADVI
            -   'svgd'  for Stein Variational Gradient Descent
            -   'asvgd'  for Amortized Stein Variational Gradient Descent
    n_samples:
        Number of samples to generate after optimization.
    random_seed: int
        random seed for reproducibility
    start_sigma: `dict[str, np.ndarray]`
        starting standard deviation for inference, only available for method 'advi'
    x0:
        Initial parameter for the variational optimization. If None, the best parameter
        found in optimization is used.
    result:
        A result to write to. If None provided, one is created from the
        problem.
    filename:
        Name of the hdf5 file, where the result will be saved. Default is
        None, which deactivates automatic saving. If set to
        "Auto" it will automatically generate a file named
        `year_month_day_profiling_result.hdf5`.
        Optionally a method, see docs for `pypesto.store.auto.autosave`.
    overwrite:
        Whether to overwrite `result/sampling` in the autosave file
        if it already exists.

    Returns
    -------
    result:
        A result with filled in sample_options part.
    """
    # prepare result object
    if result is None:
        result = Result(problem)

    # number of samples
    if n_iterations is not None:
        n_iterations = bound_n_samples_from_env(n_iterations)

    # try to find initial parameters
    if x0 is None:
        result.optimize_result.sort()
        if len(result.optimize_result.list) > 0:
            x0 = problem.get_reduced_vector(
                result.optimize_result.list[0]["x"]
            )

    # set variational inference
    # currently we only support pymc
    variational = PymcVariational()

    # initialize sampler to problem
    variational.initialize(problem=problem, x0=x0)

    # perform the sampling and track time
    t_start = process_time()
    variational.fit(
        n_iterations=n_iterations,
        method=method,
        random_seed=random_seed,
        start_sigma=start_sigma,
        **kwargs,
    )
    t_elapsed = process_time() - t_start
    logger.info("Elapsed time: " + str(t_elapsed))

    # extract results and save samples to pypesto result
    if n_samples is None or n_samples == 0:
        # constructing a McmcPtResult object with nearly empty trace_x
        n_samples = 1

    result.sample_result = variational.sample(n_samples)
    result.sample_result.time = t_elapsed

    autosave(
        filename=filename,
        result=result,
        store_type="sample",
        overwrite=overwrite,
    )

    # make pymc object available in result
    # TODO: if needed, we can add a result object for variational inference methods
    result.variational_result = variational
    (
        result.sample_result.variational_parameters_names,
        result.sample_result.variational_parameters,
    ) = variational.get_variational_parameters()
    if filename is not None:
        logger.warning(
            "Variational parameters are not saved in the hdf5 file. You have to save them manually."
        )

    return result



"""Self-adaptive cooperative enhanced scatter search (SACESS)."""

from __future__ import annotations

import itertools
import logging
import logging.handlers
import multiprocessing
import os
import time
from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from math import ceil, sqrt
from multiprocessing import get_context
from multiprocessing.managers import SyncManager
from pathlib import Path
from typing import Any
from uuid import uuid1
from warnings import warn

import numpy as np

import pypesto

from ... import MemoryHistory
from ...startpoint import StartpointMethod
from ...store.read_from_hdf5 import read_result
from ...store.save_to_hdf5 import (
    OptimizationResultHDF5Writer,
    ProblemHDF5Writer,
)
from ..optimize import Problem
from .ess import ESSExitFlag, ESSOptimizer
from .function_evaluator import create_function_evaluator
from .refset import RefSet

__all__ = [
    "SacessOptimizer",
    "get_default_ess_options",
    "SacessFidesFactory",
    "SacessCmaFactory",
    "SacessOptions",
]

logger = logging.getLogger(__name__)


class SacessOptimizer:
    """SACESS optimizer.

    A shared-memory-based implementation of the
    `Self-Adaptive Cooperative Enhanced Scatter Search` (SaCeSS) algorithm
    presented in :footcite:t:`PenasGon2017`. This is a meta-heuristic for
    global optimization. Multiple processes (`workers`) run
    :class:`enhanced scatter searches (ESSs) <ESSOptimizer>` in parallel.
    After each ESS iteration, depending on the outcome, there is a chance
    of exchanging good parameters, and changing ESS hyperparameters to those of
    the most promising worker. See :footcite:t:`PenasGon2017` for details.

    :class:`SacessOptimizer` can be used with or without a local optimizer, but
    it is highly recommended to use one.

    A basic example using :class:`SacessOptimizer` to minimize the Rosenbrock
    function:

    >>> from pypesto.optimize import SacessOptimizer
    >>> from pypesto.problem import Problem
    >>> from pypesto.objective import Objective
    >>> import scipy as sp
    >>> import numpy as np
    >>> import logging
    >>> # Define some test Problem
    >>> objective = Objective(
    ...     fun=sp.optimize.rosen,
    ...     grad=sp.optimize.rosen_der,
    ...     hess=sp.optimize.rosen_hess,
    ... )
    >>> dim = 6
    >>> problem = Problem(
    ...     objective=objective,
    ...     lb=-5 * np.ones((dim, 1)),
    ...     ub=5 * np.ones((dim, 1)),
    ... )
    >>> # Create and run the optimizer
    >>> sacess = SacessOptimizer(
    ...     num_workers=2,
    ...     max_walltime_s=5,
    ...     sacess_loglevel=logging.WARNING
    ... )
    >>> result = sacess.minimize(problem)

    .. seealso::

       :class:`pypesto.optimize.ess.ess.ESSOptimizer`

    References
    ----------
    .. footbibliography::

    Attributes
    ----------
    histories:
        List of the histories of the best values/parameters
        found by each worker. (Monotonously decreasing objective values.)
        See :func:`pypesto.visualize.optimizer_history.sacess_history` for
        visualization.
    """

    def __init__(
        self,
        num_workers: int | None = None,
        ess_init_args: list[dict[str, Any]] | None = None,
        max_walltime_s: float = np.inf,
        sacess_loglevel: int = logging.INFO,
        ess_loglevel: int = logging.WARNING,
        tmpdir: Path | str = None,
        mp_start_method: str = "spawn",
        options: SacessOptions = None,
    ):
        """Construct.

        Parameters
        ----------
        ess_init_args:
            List of argument dictionaries passed to
            :func:`ESSOptimizer.__init__`. Each entry corresponds to one worker
            process. I.e., the length of this list is the number of ESSs.
            Ideally, this list contains some more conservative and some more
            aggressive configurations.
            Resource limits such as ``max_eval`` apply to a single ESS
            iteration, not to the full search.
            Mutually exclusive with ``num_workers``.

            Recommended default settings can be obtained from
            :func:`get_default_ess_options`. For example, to run
            :class:`SacessOptimizer` without a local optimizer, use:

            >>> from pypesto.optimize.ess import get_default_ess_options
            >>> ess_init_args = get_default_ess_options(
            ...     num_workers=12,
            ...     dim=10, # usually problem.dim
            ...     local_optimizer=False,
            ... )
            >>> ess_init_args  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
            [{'dim_refset': 5, 'balance': 0.0, 'local_n1': 1, 'local_n2': 1},
            ...
             {'dim_refset': 7, 'balance': 1.0, 'local_n1': 4, 'local_n2': 4}]

        num_workers:
            Number of workers to be used. If this argument is given,
            (different) default ESS settings will be used for each worker.
            Mutually exclusive with ``ess_init_args``.
            See :func:`get_default_ess_options` for details on the default
            settings.
        max_walltime_s:
            Maximum walltime in seconds. It will only be checked between local
            optimizations and other simulations, and thus, may be exceeded by
            the duration of a local search. Defaults to no limit.
            Note that in order to impose the wall time limit also on the local
            optimizer, the user has to provide a wrapper function similar to
            :meth:`SacessFidesFactory.__call__`.
        ess_loglevel:
            Loglevel for ESS runs.
        sacess_loglevel:
            Loglevel for SACESS runs.
        tmpdir:
            Directory for temporary files. This defaults to a directory in the
            current working directory named ``SacessOptimizerTemp-{random suffix}``.
            When setting this option, make sure any optimizers running in
            parallel have a unique `tmpdir`. Expected to be empty.
        mp_start_method:
            The start method for the multiprocessing context.
            See :mod:`multiprocessing` for details. Running `SacessOptimizer`
            under Jupyter may require ``mp_start_method="fork"``.
        options:
            Further optimizer hyperparameters, see :class:`SacessOptions`.
        """
        warn(
            "Using pypesto.optimize.ess.SacessOptimizer is deprecated "
            "and will be removed in a future release. "
            "Use pyscat.SacessOptimizer instead "
            "(https://github.com/ICB-DCM/pyscat/, `pip install pyscat`).",
            DeprecationWarning,
            stacklevel=2,
        )

        if (num_workers is None and ess_init_args is None) or (
            num_workers is not None and ess_init_args is not None
        ):
            raise ValueError(
                "Exactly one of `num_workers` or `ess_init_args` "
                "has to be provided."
            )

        self.num_workers = num_workers or len(ess_init_args)
        if self.num_workers < 2:
            raise ValueError(
                f"{self.__class__.__name__} needs at least 2 workers."
            )
        self.ess_init_args = ess_init_args
        self.max_walltime_s = max_walltime_s
        self.exit_flag = ESSExitFlag.DID_NOT_RUN
        self.ess_loglevel = ess_loglevel
        self.sacess_loglevel = sacess_loglevel
        self.worker_results: list[SacessWorkerResult] = []
        logger.setLevel(self.sacess_loglevel)

        self._tmpdir = tmpdir
        if self._tmpdir is None:
            while self._tmpdir is None or self._tmpdir.exists():
                self._tmpdir = Path(f"SacessOptimizerTemp-{str(uuid1())[:8]}")
        self._tmpdir = Path(self._tmpdir).absolute()
        self._tmpdir.mkdir(parents=True, exist_ok=True)
        self.histories: list[pypesto.history.memory.MemoryHistory] | None = (
            None
        )
        self.mp_ctx = get_context(mp_start_method)
        self.options = options or SacessOptions()

    def minimize(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod = None,
    ) -> pypesto.Result:
        """Solve the given optimization problem.

        Note that if this function is called from a multithreaded program (
        multiple threads running at the time of calling this function) and
        the :mod:`multiprocessing` `start method` is set to ``fork``, there is
        a good chance for deadlocks. Postpone spawning threads until after
        `minimize` or change the *start method* to ``spawn``.

        Parameters
        ----------
        problem:
            Minimization problem.
            :meth:`Problem.startpoint_method` will be used to sample random
            points. `SacessOptimizer` will deal with non-evaluable points.
            Therefore, using :class:`pypesto.startpoint.CheckedStartpoints`
            with ``check_fval=True`` or ``check_grad=True`` is not recommended
            since it would create significant overhead.

        startpoint_method:
            Method for choosing starting points.
            **Deprecated. Use ``problem.startpoint_method`` instead.**

        Returns
        -------
        _:
            Result object with optimized parameters in
            :attr:`pypesto.Result.optimize_result`.
            Results are sorted by objective. At least the best parameters are
            included. Additional results may be included - this is subject to
            change.
        """
        if problem.x_guesses.shape[0]:
            # We'll use problem.startpoint_method to sample random points
            #  later on. Depending on the startpoint method, this will return
            #  the provided guesses, meaning that we'll always get the same
            #  points. This means, we won't explore the parameter space, or
            #  potentially even get stuck if the provided guesses are not
            #  evaluable.
            raise ValueError(
                "Providing startpoints in `problem.x_guesses` "
                f"is not supported by {self.__class__.__name__}. "
                "Unset `problem.x_guesses`."
            )

        if startpoint_method is not None:
            warn(
                "Passing `startpoint_method` directly is deprecated, "
                "use `problem.startpoint_method` instead.",
                DeprecationWarning,
                stacklevel=1,
            )

        start_time = time.time()
        logger.debug(
            f"Running {self.__class__.__name__} with {self.num_workers} "
            f"workers: {self.ess_init_args} and {self.options}."
        )
        ess_init_args = self.ess_init_args or get_default_ess_options(
            num_workers=self.num_workers, dim=problem.dim
        )

        logging_handler = logging.StreamHandler()
        logging_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(name)s %(levelname)-8s %(message)s"
            )
        )
        logging_thread = logging.handlers.QueueListener(
            self.mp_ctx.Queue(-1), logging_handler
        )

        # shared memory manager to handle shared state
        # (simulates the sacess manager process)
        with self.mp_ctx.Manager() as shmem_manager:
            sacess_manager = SacessManager(
                shmem_manager=shmem_manager,
                ess_options=ess_init_args,
                dim=problem.dim,
                options=self.options,
            )
            # create workers
            workers = [
                SacessWorker(
                    manager=sacess_manager,
                    ess_kwargs=ess_kwargs,
                    worker_idx=worker_idx,
                    max_walltime_s=self.max_walltime_s,
                    loglevel=self.sacess_loglevel,
                    ess_loglevel=self.ess_loglevel,
                    tmp_result_file=SacessWorker.get_temp_result_filename(
                        worker_idx, self._tmpdir
                    ),
                    options=self.options,
                )
                for worker_idx, ess_kwargs in enumerate(ess_init_args)
            ]
            # launch worker processes
            worker_processes = [
                self.mp_ctx.Process(
                    name=f"{self.__class__.__name__}-worker-{i:02d}",
                    target=_run_worker,
                    args=(
                        worker,
                        problem,
                        startpoint_method,
                        logging_thread.queue,
                    ),
                )
                for i, worker in enumerate(workers)
            ]
            for p in worker_processes:
                p.start()

            # start logging thread only AFTER starting the worker processes
            #  to prevent deadlocks
            logging_thread.start()

            # wait for finish
            # collect results
            self.worker_results = [
                sacess_manager._result_queue.get()
                for _ in range(self.num_workers)
            ]
            for p in worker_processes:
                p.join()

        logging_thread.stop()

        self.histories = [
            worker_result.history for worker_result in self.worker_results
        ]
        self.exit_flag = min(
            worker_result.exit_flag for worker_result in self.worker_results
        )
        result = self._create_result(problem)
        self._delete_tmpdir()

        walltime = time.time() - start_time
        n_eval_total = sum(
            worker_result.n_eval for worker_result in self.worker_results
        )
        if len(result.optimize_result):
            logger.info(
                f"{self.__class__.__name__} stopped after {walltime:3g}s "
                f"and {n_eval_total} objective evaluations "
                f"with global best {result.optimize_result[0].fval}."
            )
        else:
            logger.error(
                f"{self.__class__.__name__} stopped after {walltime:3g}s "
                f"and {n_eval_total} objective evaluations without producing "
                "a result."
            )
        return result

    def _create_result(self, problem: Problem) -> pypesto.Result:
        """Create result object.

        Creates an overall Result object from the results saved by the workers.
        """
        # gather results from workers
        result = pypesto.Result()
        retry_after_sleep = True
        for worker_idx in range(self.num_workers):
            tmp_result_filename = SacessWorker.get_temp_result_filename(
                worker_idx, self._tmpdir
            )
            tmp_result = None
            try:
                tmp_result = read_result(
                    tmp_result_filename, problem=False, optimize=True
                )
            except FileNotFoundError:
                # wait and retry, maybe the file wasn't found due to some filesystem latency issues
                if retry_after_sleep:
                    time.sleep(5)
                    # waiting once is enough - don't wait for every worker
                    retry_after_sleep = False

                    try:
                        tmp_result = read_result(
                            tmp_result_filename, problem=False, optimize=True
                        )
                    except FileNotFoundError:
                        logger.error(
                            f"Worker {worker_idx} did not produce a result."
                        )
                        continue
                else:
                    logger.error(
                        f"Worker {worker_idx} did not produce a result."
                    )
                    continue

            if tmp_result:
                result.optimize_result.append(
                    tmp_result.optimize_result,
                    sort=False,
                    prefix=f"{worker_idx}-",
                )

        result.optimize_result.sort()
        result.problem = problem

        return result

    def _delete_tmpdir(self):
        """Delete the temporary files and the temporary directory if empty."""
        for worker_idx in range(self.num_workers):
            filename = SacessWorker.get_temp_result_filename(
                worker_idx, self._tmpdir
            )
            with suppress(FileNotFoundError):
                os.remove(filename)

        # delete tmpdir if empty
        try:
            self._tmpdir.rmdir()
        except OSError:
            pass


class SacessManager:
    """The Sacess manager.

    Manages shared memory of a SACESS run. Loosely corresponds to the manager
    process in [PenasGon2017]_.

    Attributes
    ----------
    _dim: Dimension of the optimization problem
    _num_workers: Number of workers
    _ess_options: ESS options for each worker
    _best_known_fx: Best objective value encountered so far
    _best_known_x: Parameters corresponding to ``_best_known_fx``
    _worker_scores: Performance score of the different workers (the higher, the
        more promising the respective worker is considered)
    _worker_comms: Number of communications received from the individual
        workers
    _rejections: Number of rejected solutions received from workers since the
        last adaptation of ``_rejection_threshold``.
    _rejection_threshold: Threshold for relative objective improvements that
        incoming solutions have to pass to be accepted
    _lock: Lock for accessing shared state.
    _terminate: Flag to signal termination of the SACESS run to workers
    _logger: A logger instance
    _options: Further optimizer hyperparameters.
    """

    def __init__(
        self,
        shmem_manager: SyncManager,
        ess_options: list[dict[str, Any]],
        dim: int,
        options: SacessOptions = None,
    ):
        self._dim = dim
        self._options = options or SacessOptions()
        self._num_workers = len(ess_options)
        self._ess_options = [shmem_manager.dict(o) for o in ess_options]
        self._best_known_fx = shmem_manager.Value("d", np.inf)
        self._best_known_x = shmem_manager.Array("d", [np.nan] * dim)
        self._rejections = shmem_manager.Value("i", 0)
        # The initial value for the acceptance/rejection threshold in
        # [PenasGon2017]_ p.9 is 0.1.
        # However, their implementation uses 0.1 *percent*. I assume this is a
        # mistake in the paper.
        self._rejection_threshold = shmem_manager.Value(
            "d", self._options.manager_initial_rejection_threshold
        )

        # scores of the workers, ordered by worker-index
        # initial score is the worker index
        self._worker_scores = shmem_manager.Array(
            "d", range(self._num_workers)
        )
        self._terminate = shmem_manager.Value("b", False)
        self._worker_comms = shmem_manager.Array("i", [0] * self._num_workers)
        self._lock = shmem_manager.RLock()
        self._logger = logging.getLogger()
        self._result_queue = shmem_manager.Queue()

    def get_best_solution(self) -> tuple[np.ndarray, float]:
        """Get the best objective value and corresponding parameters."""
        with self._lock:
            return np.array(self._best_known_x), self._best_known_fx.value

    def reconfigure_worker(self, worker_idx: int) -> dict:
        """Reconfigure the given worker.

        Updates the ESS options for the given worker to those of the worker at
        the top of the scoreboard and returns those settings.
        """
        with self._lock:
            leader_options = self._ess_options[
                np.argmax(self._worker_scores)
            ].copy()
            for setting in ["local_n2", "balance", "dim_refset"]:
                if setting in leader_options:
                    self._ess_options[worker_idx][setting] = leader_options[
                        setting
                    ]
            return self._ess_options[worker_idx].copy()

    def submit_solution(
        self,
        x: np.ndarray,
        fx: float,
        sender_idx: int,
        elapsed_time_s: float,
    ):
        """Submit a solution.

        To be called by a worker.

        Parameters
        ----------
        x: Model parameters
        fx: Objective value corresponding to ``x``
        sender_idx: Index of the worker submitting the results.
        elapsed_time_s: Elapsed time since the beginning of the sacess run.
        """
        abs_change = fx - self._best_known_fx.value
        with self._lock:
            # cooperation step
            # solution improves best value by at least a factor of ...
            if (
                # initially _best_known_fx is NaN
                (
                    np.isfinite(fx)
                    and not np.isfinite(self._best_known_fx.value)
                )
                # avoid division by 0. just accept any improvement if the best
                # known value is 0.
                or (self._best_known_fx.value == 0 and fx < 0)
                or (
                    fx < self._best_known_fx.value
                    and abs(abs_change / self._best_known_fx.value)
                    > self._rejection_threshold.value
                )
            ):
                # accept solution
                self._logger.debug(
                    f"Accepted solution from worker {sender_idx}: {fx}."
                )
                # accept
                if len(x) != len(self._best_known_x):
                    raise AssertionError(
                        f"Received solution with {len(x)} parameters, "
                        f"but expected {len(self._best_known_x)}."
                    )
                for i, xi in enumerate(x):
                    self._best_known_x[i] = xi
                self._best_known_fx.value = fx
                self._worker_comms[sender_idx] += 1
                self._worker_scores[sender_idx] = (
                    self._worker_comms[sender_idx] * elapsed_time_s
                )
            else:
                # reject solution
                self._rejections.value += 1

                rel_change = (
                    abs(abs_change / self._best_known_fx.value)
                    if self._best_known_fx.value != 0
                    else np.nan
                )
                self._logger.debug(
                    f"Rejected solution from worker {sender_idx} "
                    f"abs change: {abs_change} "
                    f"rel change: {rel_change:.4g} "
                    f"(threshold: {self._rejection_threshold.value}) "
                    f"(total rejections: {self._rejections.value})."
                )
                # adapt the acceptance threshold if too many solutions have
                #  been rejected
                if self._rejections.value >= self._num_workers:
                    self._rejection_threshold.value = min(
                        self._rejection_threshold.value / 2,
                        self._options.manager_minimum_rejection_threshold,
                    )
                    self._logger.debug(
                        "Lowered acceptance threshold to "
                        f"{self._rejection_threshold.value}."
                    )
                    self._rejections.value = 0

    def abort(self):
        """Abort the SACESS run."""
        with self._lock:
            self._terminate.value = True

    def aborted(self) -> bool:
        """Whether this run was aborted."""
        with self._lock:
            return self._terminate.value


class SacessWorker:
    """A SACESS worker.

    Runs ESSs and exchanges information with a SacessManager.
    Corresponds to a worker process in [PenasGon2017]_.

    Attributes
    ----------
    _manager: The sacess manager this worker is working for.
    _worker_idx: Index of this worker.
    _best_known_fx: Best objective value known to this worker (obtained on its
        own or received from the manager).
    _n_received_solutions: Number of solutions received by this worker since
        the last one was sent to the manager.
    _neval: Number of objective evaluations since the last solution was sent
        to the manager.
    _ess_kwargs: ESSOptimizer options for this worker (may get updated during
        the self-adaptive step).
    _n_sent_solutions: Number of solutions sent to the Manager.
    _max_walltime_s: Walltime limit.
    _logger: A Logger instance.
    _loglevel: Logging level for sacess
    _ess_loglevel: Logging level for ESS runs
    _tmp_result_file: Path of a temporary file to be created.
    """

    def __init__(
        self,
        manager: SacessManager,
        ess_kwargs: dict[str, Any],
        worker_idx: int,
        max_walltime_s: float = np.inf,
        loglevel: int = logging.INFO,
        ess_loglevel: int = logging.WARNING,
        tmp_result_file: str = None,
        options: SacessOptions = None,
    ):
        self._manager = manager
        self._worker_idx = worker_idx
        self._best_known_fx = np.inf
        self._n_received_solutions = 0
        self._neval = 0
        self._ess_kwargs = ess_kwargs
        self._n_sent_solutions = 0
        self._max_walltime_s = max_walltime_s
        self._start_time = None
        self._loglevel = loglevel
        self._ess_loglevel = ess_loglevel
        self._logger = None
        self._tmp_result_file = tmp_result_file
        self._refset = None
        self._options = options or SacessOptions()

    def run(
        self,
        problem: Problem,
        startpoint_method: StartpointMethod,
    ):
        self._start_time = time.time()

        # index of the local solution in ESSOptimizer.local_solutions
        #  that was most recently saved by _autosave
        last_saved_local_solution = -1

        self._logger.setLevel(self._loglevel)
        # Set the manager logger to one created within the current process
        self._manager._logger = self._logger

        self._logger.debug(
            f"#{self._worker_idx} starting "
            f"({self._ess_kwargs}, {self._options})."
        )

        evaluator = create_function_evaluator(
            problem,
            startpoint_method,
            n_procs=self._ess_kwargs.get("n_procs"),
            n_threads=self._ess_kwargs.get("n_threads"),
        )

        # create initial refset
        self._refset = RefSet(
            dim=self._ess_kwargs["dim_refset"], evaluator=evaluator
        )
        self._refset.initialize_random(
            n_diverse=max(
                self._ess_kwargs.get("n_diverse", 10 * problem.dim),
                self._refset.dim,
            )
        )

        ess = self._setup_ess(startpoint_method)

        # run ESS until exit criteria are met, but start at least one iteration
        while self._keep_going(ess) or ess.n_iter == 0:
            # perform one ESS iteration
            ess._do_iteration()

            # check if the best solution of the last local ESS is sufficiently
            # better than the sacess-wide best solution
            self.maybe_update_best(ess.x_best, ess.fx_best)
            self._best_known_fx = min(ess.fx_best, self._best_known_fx)

            self._autosave(ess, last_saved_local_solution)
            last_saved_local_solution = len(ess.local_solutions) - 1

            self._cooperate()
            self._maybe_adapt(problem)

            t_left = self._max_walltime_s - (time.time() - self._start_time)
            self._logger.info(
                f"sacess worker {self._worker_idx} iteration {ess.n_iter} "
                f"(best: {self._best_known_fx}, "
                f"n_eval: {ess.evaluator.n_eval}, "
                f"remaining wall time: {t_left}s)."
            )
        self._finalize(ess)

    def _finalize(self, ess: ESSOptimizer = None):
        """Finalize the worker."""
        # Whatever happens here, we need to put something to the queue before
        # returning to avoid deadlocks.
        worker_result = None
        if ess is not None:
            try:
                ess.history.finalize(exitflag=ess.exit_flag.name)
                ess._report_final()
                worker_result = SacessWorkerResult(
                    x=ess.x_best,
                    fx=ess.fx_best,
                    history=ess.history,
                    n_eval=ess.evaluator.n_eval,
                    n_iter=ess.n_iter,
                    n_local=len(ess.local_solutions),
                    exit_flag=ess.exit_flag,
                )
            except Exception as e:
                self._logger.exception(
                    f"Worker {self._worker_idx} failed to finalize: {e}"
                )
        if worker_result is None:
            # Create some dummy result
            worker_result = SacessWorkerResult(
                x=np.full(self._manager._dim, np.nan),
                fx=np.nan,
                history=MemoryHistory(),
                n_eval=0,
                n_iter=0,
                n_local=0,
                exit_flag=ESSExitFlag.ERROR,
            )
        self._manager._result_queue.put(worker_result)

        self._logger.debug(f"Final configuration: {self._ess_kwargs}")

    def _setup_ess(self, startpoint_method: StartpointMethod) -> ESSOptimizer:
        """Run ESS."""
        ess_kwargs = self._ess_kwargs.copy()
        # account for sacess walltime limit
        ess_kwargs["max_walltime_s"] = min(
            ess_kwargs.get("max_walltime_s", np.inf),
            self._max_walltime_s - (time.time() - self._start_time),
        )

        ess = ESSOptimizer(**ess_kwargs)
        ess.logger = self._logger.getChild(
            f"sacess-{self._worker_idx:02d}-ess"
        )
        ess.logger.setLevel(self._ess_loglevel)

        ess._initialize_minimize(
            startpoint_method=startpoint_method, refset=self._refset
        )

        return ess

    def _cooperate(self):
        """Cooperation step."""
        # try to obtain a new solution from manager
        recv_x, recv_fx = self._manager.get_best_solution()
        self._logger.log(
            logging.DEBUG - 1,
            f"Worker {self._worker_idx} received solution {recv_fx} "
            f"(known best: {self._best_known_fx}).",
        )
        if recv_fx < self._best_known_fx or (
            not np.isfinite(self._best_known_fx) and np.isfinite(recv_fx)
        ):
            if not np.isfinite(recv_x).all():
                raise AssertionError(
                    f"Received non-finite parameters {recv_x}."
                )
            self._logger.debug(
                f"Worker {self._worker_idx} received better solution {recv_fx}."
            )
            self._best_known_fx = recv_fx
            self._n_received_solutions += 1
            self.replace_solution(self._refset, x=recv_x, fx=recv_fx)

    def _maybe_adapt(self, problem: Problem):
        """Perform the adaptation step if needed.

        Update ESS settings if conditions are met.
        """
        # Update ESS settings if we received way more solutions than we sent
        #  Note: [PenasGon2017]_ Algorithm 5 uses AND in the following
        #  condition, but the accompanying implementation uses OR.
        if (
            self._n_received_solutions
            > self._options.adaptation_sent_coeff * self._n_sent_solutions
            + self._options.adaptation_sent_offset
            or self._neval > problem.dim * self._options.adaptation_min_evals
        ):
            self._ess_kwargs = self._manager.reconfigure_worker(
                self._worker_idx
            )
            self._refset.sort()
            self._refset.resize(self._ess_kwargs["dim_refset"])
            self._logger.debug(
                f"Updated settings on worker {self._worker_idx} to "
                f"{self._ess_kwargs}"
            )
        else:
            self._logger.debug(
                f"Worker {self._worker_idx} not adapting. "
                f"Received: {self._n_received_solutions} <= {self._options.adaptation_sent_coeff * self._n_sent_solutions + self._options.adaptation_sent_offset}, "
                f"Sent: {self._n_sent_solutions}, "
                f"neval: {self._neval} <= {problem.dim * self._options.adaptation_min_evals}."
            )

    def maybe_update_best(self, x: np.ndarray, fx: float):
        """Maybe update the best known solution and send it to the manager."""
        rel_change = (
            abs((fx - self._best_known_fx) / fx) if fx != 0 else np.nan
        )
        self._logger.debug(
            f"Worker {self._worker_idx} maybe sending solution {fx}. "
            f"best known: {self._best_known_fx}, "
            f"rel change: {rel_change:.4g}, "
            f"threshold: {self._options.worker_acceptance_threshold}"
        )

        # solution improves the best value by at least a factor of ...
        if (
            (np.isfinite(fx) and not np.isfinite(self._best_known_fx))
            or (self._best_known_fx == 0 and fx < 0)
            or (
                fx < self._best_known_fx
                and abs((self._best_known_fx - fx) / self._best_known_fx)
                > self._options.worker_acceptance_threshold
            )
        ):
            self._logger.debug(
                f"Worker {self._worker_idx} sending solution {fx}."
            )
            self._n_sent_solutions += 1
            self._best_known_fx = fx

            # send best_known_fx to manager
            self._neval = 0
            self._n_received_solutions = 0
            elapsed_time = time.time() - self._start_time
            self._manager.submit_solution(
                x=x,
                fx=fx,
                sender_idx=self._worker_idx,
                elapsed_time_s=elapsed_time,
            )

    @staticmethod
    def replace_solution(refset: RefSet, x: np.ndarray, fx: float):
        """Replace the global refset member by the given solution."""
        # [PenasGon2017]_ page 8, top
        if "cooperative_solution" not in refset.attributes:
            label = np.zeros(shape=refset.dim)
            # on first call, mark the worst solution as "cooperative solution"
            cooperative_solution_idx = np.argmax(refset.fx)
            label[cooperative_solution_idx] = 1
            refset.add_attribute("cooperative_solution", label)
        elif (
            cooperative_solution_idx := np.argwhere(
                refset.attributes["cooperative_solution"]
            )
        ).size == 0:
            # the attribute exists, but no member is marked as the cooperative
            # solution. this may happen if we shrink the refset.
            cooperative_solution_idx = np.argmax(refset.fx)

        # replace the cooperative solution
        refset.update(
            i=cooperative_solution_idx,
            x=x,
            fx=fx,
        )

    def _keep_going(self, ess: ESSOptimizer) -> bool:
        """Check exit criteria.

        Returns
        -------
        ``True`` if none of the exit criteria is met, ``False`` otherwise.
        """
        # elapsed time
        if time.time() - self._start_time >= self._max_walltime_s:
            ess.exit_flag = ESSExitFlag.MAX_TIME
            self._logger.debug(
                f"Max walltime ({self._max_walltime_s}s) exceeded."
            )
            return False
        # other reasons for termination (some worker failed, ...)
        if self._manager.aborted():
            ess.exit_flag = ESSExitFlag.ERROR
            self._logger.debug("Manager requested termination.")
            return False
        return True

    def abort(self):
        """Send signal to abort."""
        self._logger.error(f"Worker {self._worker_idx} aborting.")
        # signal to manager
        self._manager.abort()

        self._finalize(None)

    def _autosave(self, ess: ESSOptimizer, last_saved_local_solution: int):
        """Save intermediate results.

        If a temporary result file is set, save the (part of) the current state
        of the ESS to that file.

        We save the current best solution and the local optimizer results.
        """
        if not self._tmp_result_file:
            return

        t_start = time.time()

        # save problem in first iteration
        if ess.n_iter == 1:
            pypesto_problem_writer = ProblemHDF5Writer(self._tmp_result_file)
            pypesto_problem_writer.write(
                ess.evaluator.problem, overwrite=False
            )

        opt_res_writer = OptimizationResultHDF5Writer(self._tmp_result_file)
        for i in range(
            last_saved_local_solution + 1, len(ess.local_solutions)
        ):
            optimizer_result = ess.local_solutions[i]
            optimizer_result.id = str(i + ess.n_iter)
            opt_res_writer.write_optimizer_result(
                optimizer_result, overwrite=False
            )

        # save the current best solution
        optimizer_result = pypesto.OptimizerResult(
            id=str(len(ess.local_solutions) + ess.n_iter),
            x=ess.x_best,
            fval=ess.fx_best,
            message=f"Global best (iteration {ess.n_iter})",
            time=time.time() - ess.starttime,
            n_fval=ess.evaluator.n_eval,
            optimizer=str(ess),
        )
        optimizer_result.update_to_full(ess.evaluator.problem)
        opt_res_writer.write_optimizer_result(
            optimizer_result, overwrite=False
        )

        t_save = time.time() - t_start
        self._logger.debug(
            f"Worker {self._worker_idx} autosave to {self._tmp_result_file} "
            f"took {t_save:.2f}s."
        )

    @staticmethod
    def get_temp_result_filename(worker_idx: int, tmpdir: str | Path) -> str:
        return str(Path(tmpdir, f"sacess-{worker_idx:02d}_tmp.h5").absolute())


def _run_worker(
    worker: SacessWorker,
    problem: Problem,
    startpoint_method: StartpointMethod,
    log_process_queue: multiprocessing.Queue,
):
    """Run the given SACESS worker.

    Helper function as entrypoint for sacess worker processes.
    """
    try:
        # different random seeds per process
        np.random.seed((os.getpid() * int(time.time() * 1000)) % 2**32)

        # Forward log messages to the logging process
        h = logging.handlers.QueueHandler(log_process_queue)
        worker._logger = logging.getLogger(
            multiprocessing.current_process().name
        )
        worker._logger.addHandler(h)

        return worker.run(problem=problem, startpoint_method=startpoint_method)
    except Exception as e:
        with suppress(Exception):
            worker._logger.exception(
                f"Worker {worker._worker_idx} failed: {e}"
            )
        worker.abort()


def get_default_ess_options(
    num_workers: int,
    dim: int,
    local_optimizer: bool
    | pypesto.optimize.Optimizer
    | Callable[..., pypesto.optimize.Optimizer] = True,
) -> list[dict]:
    """Get default ESS settings for (SA)CESS.

    Returns settings for ``num_workers`` parallel scatter searches, combining
    more aggressive and more conservative configurations. Mainly intended for
    use with :class:`SacessOptimizer`. For details on the different options,
    see keyword arguments of :meth:`ESSOptimizer.__init__`.

    Setting appropriate values for ``n_threads`` and ``local_optimizer`` is
    left to the user. Defaults to single-threaded and no local optimizer.

    Based on https://bitbucket.org/DavidPenas/sacess-library/src/508e7ac15579104731cf1f8c3969960c6e72b872/src/method_module_fortran/eSS/parallelscattersearchfunctions.f90#lines-929

    Parameters
    ----------
    num_workers: Number of configurations to return.
    dim: Problem dimension (number of optimized parameters).
    local_optimizer: The local optimizer to use
        (see same argument in :class:`ESSOptimizer`), a boolean indicating
        whether to set the default local optimizer
        (currently :class:`FidesOptimizer`), a :class:`Optimizer` instance,
        or a :obj:`Callable` returning an optimizer instance.
        The latter can be used to propagate walltime limits to the local
        optimizers. See :meth:`SacessFidesFactory.__call__` for an example.
        The current default optimizer assumes that the optimized objective
        function can provide its gradient. If this is not the case, the user
        should provide a different local optimizer or consider using
        :class:`pypesto.objective.finite_difference.FD` to approximate the
        gradient using finite differences.
    """
    min_dimrefset = 5

    def dim_refset(x):
        return max(min_dimrefset, ceil((1 + sqrt(4 * dim * x)) / 2))

    settings = [
        # 1
        {
            "dim_refset": dim_refset(1),
            "balance": 0.0,
            "local_n1": 1,
            "local_n2": 1,
        },
        # 2
        {
            "dim_refset": dim_refset(3),
            "balance": 0.0,
            "local_n1": 1000,
            "local_n2": 1000,
        },
        # 3
        {
            "dim_refset": dim_refset(5),
            "balance": 0.25,
            "local_n1": 10,
            "local_n2": 10,
        },
        # 4
        {
            "dim_refset": dim_refset(10),
            "balance": 0.5,
            "local_n1": 20,
            "local_n2": 20,
        },
        # 5
        {
            "dim_refset": dim_refset(15),
            "balance": 0.25,
            "local_n1": 100,
            "local_n2": 100,
        },
        # 6
        {
            "dim_refset": dim_refset(12),
            "balance": 0.25,
            "local_n1": 1000,
            "local_n2": 1000,
        },
        # 7
        {
            "dim_refset": dim_refset(7.5),
            "balance": 0.25,
            "local_n1": 15,
            "local_n2": 15,
        },
        # 8
        {
            "dim_refset": dim_refset(5),
            "balance": 0.25,
            "local_n1": 7,
            "local_n2": 7,
        },
        # 9
        {
            "dim_refset": dim_refset(2),
            "balance": 0.0,
            "local_n1": 1000,
            "local_n2": 1000,
        },
        # 10
        {
            "dim_refset": dim_refset(0.5),
            "balance": 0.0,
            "local_n1": 1,
            "local_n2": 1,
        },
        # 11
        {
            "dim_refset": dim_refset(1.5),
            "balance": 1.0,
            "local_n1": 1,
            "local_n2": 1,
        },
        # 12
        {
            "dim_refset": dim_refset(3.5),
            "balance": 1.0,
            "local_n1": 4,
            "local_n2": 4,
        },
        # 13
        {
            "dim_refset": dim_refset(5.5),
            "balance": 0.1,
            "local_n1": 10,
            "local_n2": 10,
        },
        # 14
        {
            "dim_refset": dim_refset(10.5),
            "balance": 0.3,
            "local_n1": 20,
            "local_n2": 20,
        },
        # 15
        {
            "dim_refset": dim_refset(15.5),
            "balance": 0.2,
            "local_n1": 1000,
            "local_n2": 1000,
        },
        # 16
        {
            "dim_refset": dim_refset(12.5),
            "balance": 0.2,
            "local_n1": 10,
            "local_n2": 10,
        },
        # 17
        {
            "dim_refset": dim_refset(8),
            "balance": 0.75,
            "local_n1": 15,
            "local_n2": 15,
        },
        # 18
        {
            "dim_refset": dim_refset(5.5),
            "balance": 0.75,
            "local_n1": 1000,
            "local_n2": 1000,
        },
        # 19
        {
            "dim_refset": dim_refset(2.2),
            "balance": 1.0,
            "local_n1": 2,
            "local_n2": 2,
        },
        # 20
        {
            "dim_refset": dim_refset(1),
            "balance": 1.0,
            "local_n1": 1,
            "local_n2": 1,
        },
    ]

    # Set local optimizer
    for cur_settings in settings:
        if local_optimizer is True:
            cur_settings["local_optimizer"] = SacessFidesFactory(
                fides_kwargs={"verbose": logging.WARNING}
            )
        elif local_optimizer is not False:
            cur_settings["local_optimizer"] = local_optimizer

    return list(itertools.islice(itertools.cycle(settings), num_workers))


class SacessFidesFactory:
    """Factory for :class:`FidesOptimizer` instances for use with :class:`SacessOptimizer`.

    :meth:`__call__` will forward the walltime limit and function evaluation
    limit imposed on :class:`SacessOptimizer` to :class:`FidesOptimizer`.
    Besides that, default options are used.


    Parameters
    ----------
    fides_options:
        Options for the :class:`FidesOptimizer`.
        See :class:`fides.constants.Options`.
    fides_kwargs:
        Keyword arguments for the :class:`FidesOptimizer`. See
        :meth:`FidesOptimizer.__init__`. Must not include ``options``.
    """

    def __init__(
        self,
        fides_options: dict[str, Any] | None = None,
        fides_kwargs: dict[str, Any] | None = None,
    ):
        if fides_options is None:
            fides_options = {}
        if fides_kwargs is None:
            fides_kwargs = {}

        self._fides_options = fides_options
        self._fides_kwargs = fides_kwargs

        # Check if fides is installed
        try:
            import fides  # noqa F401
        except ImportError:
            from ..optimizer import OptimizerImportError

            raise OptimizerImportError("fides") from None

    def __call__(
        self, max_walltime_s: int, max_eval: int
    ) -> pypesto.optimize.FidesOptimizer:
        """Create a :class:`FidesOptimizer` instance."""

        from fides.constants import Options as FidesOptions

        options = self._fides_options.copy()
        options[FidesOptions.MAXTIME] = max_walltime_s

        # only accepts int
        if np.isfinite(max_eval):
            options[FidesOptions.MAXITER] = int(max_eval)
        return pypesto.optimize.FidesOptimizer(
            **self._fides_kwargs, options=options
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(fides_options={self._fides_options}, fides_kwargs={self._fides_kwargs})"


class SacessCmaFactory:
    """Factory for :class:`CmaOptimizer` instances for use with :class:`SacessOptimizer`.

    :meth:`__call__` will forward the walltime limit and function evaluation
    limit imposed on :class:`SacessOptimizer` to :class:`CmaOptimizer`.
    Besides that, default options are used.


    Parameters
    ----------
    options:
        Options as passed to :meth:`CmaOptimizer.__init__`.
        See ``cma.CMAOptions()`` for available options.
    """

    def __init__(
        self,
        options: dict[str, Any] | None = None,
    ):
        if options is None:
            options = {}

        self._options = options

        # Check if cma is installed
        try:
            import cma  # noqa F401
        except ImportError:
            from ..optimizer import OptimizerImportError

            raise OptimizerImportError("cma") from None

    def __call__(
        self, max_walltime_s: int, max_eval: int
    ) -> pypesto.optimize.CmaOptimizer:
        """Create a :class:`CmaOptimizer` instance."""
        options = self._options.copy()
        options["timeout"] = max_walltime_s
        options["maxfevals"] = max_eval

        return pypesto.optimize.CmaOptimizer(options=options)

    def __repr__(self):
        return f"{self.__class__.__name__}(options={self._options})"


class SacessIpoptFactory:
    """Factory for :class:`IpoptOptimizer` instances for use with :class:`SacessOptimizer`.

    :meth:`__call__` will forward the walltime limit and function evaluation
    limit imposed on :class:`SacessOptimizer` to :class:`IpoptOptimizer`.
    Besides that, default options are used.


    Parameters
    ----------
    ipopt_options:
        Options for the :class:`IpoptOptimizer`.
        See https://coin-or.github.io/Ipopt/OPTIONS.html.
    """

    def __init__(
        self,
        ipopt_options: dict[str, Any] | None = None,
    ):
        if ipopt_options is None:
            ipopt_options = {}

        self._ipopt_options = ipopt_options

        import cyipopt

        if cyipopt.IPOPT_VERSION < (3, 14, 0):
            ver = ".".join(map(str, cyipopt.IPOPT_VERSION))
            warn(
                f"The currently installed Ipopt version {ver} "
                "does not support the `max_wall_time` option. "
                "At least Ipopt 3.14 is required. "
                "The walltime limit will be ignored.",
                stacklevel=2,
            )

    def __call__(
        self, max_walltime_s: float, max_eval: float
    ) -> pypesto.optimize.IpoptOptimizer:
        """Create a :class:`IpoptOptimizer` instance."""
        import cyipopt

        options = self._ipopt_options.copy()
        if np.isfinite(max_walltime_s) and cyipopt.IPOPT_VERSION >= (3, 14, 0):
            options["max_wall_time"] = max_walltime_s

        if np.isfinite(max_eval):
            raise NotImplementedError(
                "Ipopt does not support function evaluation limits."
            )
        return pypesto.optimize.IpoptOptimizer(options=options)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(ipopt_options={self._ipopt_options})"
        )


@dataclass
class SacessWorkerResult:
    """Container for :class:`SacessWorker` results.

    Contains various information about the optimization process of a single
    :class:`SacessWorker` instance that is to be sent to
    :class:`SacessOptimizer`.

    Attributes
    ----------
    x:
        Best parameters found.
    fx:
        Objective value corresponding to ``x``.
    n_eval:
        Number of objective evaluations performed.
    n_iter:
        Number of scatter search iterations performed.
    n_local:
        Number of local searches performed.
    history:
        History object containing information about the optimization process.
    exit_flag:
        Exit flag of the optimization process.
    """

    x: np.ndarray
    fx: float
    n_eval: int
    n_iter: int
    n_local: int
    history: pypesto.history.memory.MemoryHistory
    exit_flag: ESSExitFlag


@dataclass
class SacessOptions:
    """Container for :class:`SacessOptimizer` hyperparameters.

    Parameters
    ----------
    manager_initial_rejection_threshold, manager_minimum_rejection_threshold:
        Initial and minimum threshold for relative objective improvements that
        incoming solutions have to pass to be accepted. If the number of
        rejected solutions exceeds the number of workers, the threshold is
        halved until it reaches ``manager_minimum_rejection_threshold``.

    worker_acceptance_threshold:
        Minimum relative improvement of the objective compared to the best
        known value to be eligible for submission to the Manager.

    adaptation_min_evals, adaptation_sent_offset, adaptation_sent_coeff:
        Hyperparameters that control when the workers will adapt their settings
        based on the performance of the other workers.

        The adaptation step is performed if all the following conditions are
        met:

        * The number of function evaluations since the last solution was sent
          to the manager times the number of optimization parameters is greater
          than ``adaptation_min_evals``.

        * The number of solutions received by the worker since the last
          solution it sent to the manager is greater than
          ``adaptation_sent_coeff * n_sent_solutions + adaptation_sent_offset``,
          where ``n_sent_solutions`` is the number of solutions sent to the
          manager by the given worker.

    """

    manager_initial_rejection_threshold: float = 0.001
    manager_minimum_rejection_threshold: float = 0.001

    # Default value from original SaCeSS implementation
    worker_acceptance_threshold: float = 0.0001

    # Magic numbers for adaptation, taken from [PenasGon2017]_ algorithm 5
    adaptation_min_evals: int = 5000
    adaptation_sent_offset: int = 20
    adaptation_sent_coeff: int = 10

    def __post_init__(self):
        if self.adaptation_min_evals < 0:
            raise ValueError("adaptation_min_evals must be non-negative.")
        if self.adaptation_sent_offset < 0:
            raise ValueError("adaptation_sent_offset must be non-negative.")
        if self.adaptation_sent_coeff < 0:
            raise ValueError("adaptation_sent_coeff must be non-negative.")
        if self.manager_initial_rejection_threshold < 0:
            raise ValueError(
                "manager_initial_rejection_threshold must be non-negative."
            )
        if self.manager_minimum_rejection_threshold < 0:
            raise ValueError(
                "manager_minimum_rejection_threshold must be non-negative."
            )
        if self.worker_acceptance_threshold < 0:
            raise ValueError(
                "worker_acceptance_threshold must be non-negative."
            )



"""Utility functions for working with roadrunner and PEtab.

Includes ExpData class for managing experimental data, SolverOptions class for
managing roadrunner solver options, and utility functions to convert between
PEtab measurement data and a roarunner simulation output back and forth.
"""

from __future__ import annotations

import warnings
from collections.abc import Sequence

import numpy as np
import pandas as pd

try:
    import petab.v1 as petab
    from petab.v1.C import (
        LIN,
        MEASUREMENT,
        NOISE_DISTRIBUTION,
        NOISE_FORMULA,
        NORMAL,
        OBSERVABLE_ID,
        OBSERVABLE_TRANSFORMATION,
        SIMULATION,
        SIMULATION_CONDITION_ID,
        TIME,
    )
except ImportError:
    petab = None

try:
    import roadrunner
except ImportError:
    roadrunner = None


class ExpData:
    """Class for managing experimental data for a single condition."""

    def __init__(
        self,
        condition_id: str,
        measurements: np.ndarray,
        observable_ids: Sequence[str],
        noise_distributions: np.ndarray,
        noise_formulae: np.ndarray,
    ):
        """
        Initialize the ExpData object.

        Parameters
        ----------
        condition_id:
            Identifier of the condition.
        measurements:
            Numpy Array containing the measurement data. It is a 2D array of
            dimension (n_timepoints, n_observables + 1). The first column is
            the timepoints, the remaining columns are the observable values.
            Observables not measured at a given timepoint should be NaN.
        timepoints:
            Timepoints of the measurement data.
        observable_ids:
            Observable ids of the measurement data. Order must match the
            columns of the measurements array (-time).
        noise_distributions:
            Numpy Array describing noise distributions of the measurement
            data. Dimension: (n_timepoints, n_observables). Each entry is a
            string describing scale and type of noise distribution, the name
            is "scale_type". E.g. "lin_normal", "log_normal", "log10_normal".
        noise_formulae:
            Numpy Array describing noise formulae of the measurement data.
            Dimension: (n_timepoints, n_observables). Each entry is a string
            describing the noise formula, either a parameter name or a constant.
        """
        self.condition_id = condition_id
        self.measurements = measurements
        self.observable_ids = observable_ids
        self.noise_distributions = noise_distributions
        self.noise_formulae = noise_formulae

        self.sanity_check()

    # define timepoints as a property
    @property
    def timepoints(self):
        """Timepoints of the measurement data."""
        return self.measurements[:, 0]

    def get_observable_ids(self):
        """
        Get the observable ids of the measurement data.

        Returns
        -------
        observable_ids:
            Observable ids of the measurement data.
        """
        return self.observable_ids

    def sanity_check(self):
        """Perform a sanity check of the data."""
        if self.measurements.shape[1] != len(self.observable_ids) + 1:
            raise ValueError(
                "Number of columns in measurements does not match number of "
                "observable ids + time."
            )
        # check that the noise distributions and noise formulae have the
        # same length as the number of observables
        if len(self.noise_distributions) != len(self.observable_ids):
            raise ValueError(
                "Number of noise distributions does not match number of "
                "observable ids."
            )
        if len(self.noise_formulae) != len(self.observable_ids):
            raise ValueError(
                "Number of noise formulae does not match number of "
                "observable ids."
            )

    @staticmethod
    def from_petab_problem(petab_problem: petab.Problem) -> list[ExpData]:
        """
        Create a list of ExpData object from a petab problem.

        Parameters
        ----------
        petab_problem:
            PEtab problem.
        """
        # extract all condition ids from measurement data
        condition_ids = list(
            petab_problem.measurement_df["simulationConditionId"].unique()
        )
        exp_datas = [
            ExpData.from_petab_single_condition(
                condition_id=condition_id, petab_problem=petab_problem
            )
            for condition_id in condition_ids
        ]
        return exp_datas

    @staticmethod
    def from_petab_single_condition(
        condition_id: str, petab_problem: petab.Problem
    ) -> ExpData:
        """
        Create an ExpData object from a single condition of a petab problem.

        Parameters
        ----------
        condition_id:
            Identifier of the condition.
        petab_problem:
            PEtab problem.
        """
        # extract measurement data for a single condition
        measurement_df = petab_problem.measurement_df[
            petab_problem.measurement_df[SIMULATION_CONDITION_ID]
            == condition_id
        ]
        # turn measurement data into a numpy array
        measurements, observale_ids = measurement_df_to_matrix(measurement_df)
        # construct noise distributions and noise formulae
        noise_distributions, noise_formulae = construct_noise_matrices(
            petab_problem, observale_ids
        )
        return ExpData(
            condition_id=condition_id,
            measurements=measurements,
            observable_ids=observale_ids,
            noise_distributions=noise_distributions,
            noise_formulae=noise_formulae,
        )


class SolverOptions(dict):
    """Class for managing solver options of roadrunner."""

    def __init__(
        self,
        integrator: str | None = None,
        relative_tolerance: float | None = None,
        absolute_tolerance: float | None = None,
        maximum_num_steps: int | None = None,
        **kwargs,
    ):
        """
        Initialize the SolverOptions object. Can be used as a dictionary.

        Parameters
        ----------
        integrator:
            Integrator to use.
        relative_tolerance:
            Relative tolerance of the integrator.
        absolute_tolerance:
            Absolute tolerance of the integrator.
        maximum_num_steps:
            Maximum number of steps to take.
        kwargs:
            Additional solver options.
        """
        super().__init__()
        if integrator is None:
            integrator = "cvode"
        self.integrator = integrator
        if relative_tolerance is None:
            relative_tolerance = 1e-6
        self.relative_tolerance = relative_tolerance
        if absolute_tolerance is None:
            absolute_tolerance = 1e-12
        self.absolute_tolerance = absolute_tolerance
        if maximum_num_steps is None:
            maximum_num_steps = 20000
        self.maximum_num_steps = maximum_num_steps
        self.update(kwargs)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key) from None

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        """Return a dict representation of the SolverOptions object."""
        return f"{self.__class__.__name__}({super().__repr__()})"

    def apply_to_roadrunner(self, roadrunner_instance: roadrunner.RoadRunner):
        """
        Apply the solver options to a roadrunner object inplace.

        Parameters
        ----------
        roadrunner_instance:
            Roadrunner object to apply the solver options to.
        """
        # don't allow 'gillespie' integrator
        if self.integrator == "gillespie":
            raise ValueError("Gillespie integrator is not supported.")
        # copy the options
        options = self.copy()
        # set integrator and remove integrator from options
        roadrunner_instance.setIntegrator(options.pop("integrator"))
        integrator = roadrunner_instance.getIntegrator()
        # set the remaining options
        for key, value in options.items():
            # try to set the options, if it fails, raise a warning
            try:
                integrator.setValue(key, value)
            except RuntimeError as e:
                warnings.warn(
                    f"Failed to set option {key} to {value}. Reason: {e}. "
                    f"Valid keys are: {integrator.getSettings()}.",
                    stacklevel=2,
                )


def unscale_parameters(value_dict: dict, petab_scale_dict: dict) -> dict:
    """
    Unscale the scaled parameters from target scale to linear.

    Parameters
    ----------
    value_dict:
        Dictionary with values to scale.
    petab_scale_dict:
        Target Scales.

    Returns
    -------
    unscaled_parameters:
        Dict of unscaled parameters.
    """
    if value_dict.keys() != petab_scale_dict.keys():
        raise AssertionError("Keys don't match.")

    for key, value in value_dict.items():
        value_dict[key] = petab.parameters.unscale(
            value, petab_scale_dict[key]
        )

    return value_dict


def measurement_df_to_matrix(
    measurement_df: pd.DataFrame,
) -> tuple[np.ndarray, list[str]]:
    """
    Convert a PEtab measurement DataFrame to a matrix.

    Parameters
    ----------
    measurement_df:
        DataFrame containing the measurement data.

    Returns
    -------
    measurement_matrix:
        Numpy array containing the measurement data. It is a 2D array of
        dimension (n_timepoints, n_observables + 1). The first column is
        the timepoints, the remaining columns are the observable values.
        Observables not measured at a given timepoint will be NaN.
    observable_ids:
        Observable ids of the measurement data.
    """
    measurement_df = measurement_df.loc[
        :, ["observableId", "time", "measurement"]
    ]
    # get potential replicates via placeholder "count"
    measurement_df["count"] = measurement_df.groupby(
        ["observableId", "time"]
    ).cumcount()
    pivot_df = measurement_df.pivot(
        index=["time", "count"],
        columns="observableId",
        values="measurement",
    ).fillna(np.nan)
    pivot_df.reset_index(inplace=True)
    pivot_df.drop(columns="count", inplace=True)

    observable_ids = pivot_df.columns[1:]
    measurement_matrix = pivot_df.to_numpy()

    return measurement_matrix, list(observable_ids)


def construct_noise_matrices(
    petab_problem: petab.Problem,
    observable_ids: Sequence[str],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Construct noise matrices from a PEtab problem.

    Parameters
    ----------
    petab_problem:
        PEtab problem.
    observable_ids:
        Observable ids of the measurement data.

    Returns
    -------
    noise_distributions:
        Numpy Array describing noise distributions of the measurement
        data. Dimension: (1, n_observables). Each entry is a
        string describing scale and type of noise distribution, the name
        is "scale_type". E.g. "lin_normal", "log_normal", "log10_normal".
    noise_formulae:
        Numpy Array describing noise formulae of the measurement data.
        Dimension: (1, n_observables). Each entry is a string
        describing the noise formula, either a parameter name or a constant.
    """

    def _get_noise(observable_id: str) -> tuple[str, str]:
        """
        Get noise distribution and noise formula for a single observable.

        Parameters
        ----------
        observable_id:
            Identifier of the observable.

        Returns
        -------
        noise:
            Tuple of noise distribution and noise formula.
        """
        obs_df = petab_problem.observable_df
        # check whether Index name is "observableId", if yes, get the row
        if obs_df.index.name == OBSERVABLE_ID:
            row = obs_df.loc[observable_id]
        elif OBSERVABLE_ID in obs_df.columns:
            row = obs_df[obs_df[OBSERVABLE_ID] == observable_id].iloc[0]
        else:
            raise ValueError("No observableId in observable_df.")
        # noise distribution
        noise_scale = LIN
        noise_type = NORMAL
        # check if "observableTransformation" and "noiseDistribution" exist
        if OBSERVABLE_TRANSFORMATION in obs_df.columns:
            if not pd.isna(row[OBSERVABLE_TRANSFORMATION]):
                noise_scale = row[OBSERVABLE_TRANSFORMATION]
        if NOISE_DISTRIBUTION in obs_df.columns:
            if not pd.isna(row[NOISE_DISTRIBUTION]):
                noise_type = row[NOISE_DISTRIBUTION]
        noise_distribution = f"{noise_scale}_{noise_type}"
        # TODO: check if noise_distribution is a allowed
        # noise formula
        noise_formula = row[NOISE_FORMULA]
        return noise_distribution, noise_formula

    # extract noise distributions and noise formulae
    noise = [_get_noise(observable_id) for observable_id in observable_ids]

    noise_distributions, noise_formulae = zip(*noise, strict=True)
    return np.array(noise_distributions), np.array(noise_formulae)


def simulation_to_measurement_df(
    simulations: dict,
    measurement_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Convert simulation results to a measurement DataFrame.

    Parameters
    ----------
    simulations:
        Dictionary containing the simulation results of a roadrunner
        simulator. The keys are the condition ids and the values are the
        simulation results.
    measurement_df:
        DataFrame containing the measurement data of the PEtab problem.
    """
    simulation_conditions = petab.get_simulation_conditions(measurement_df)
    meas_dfs = []
    for _, condition_id in simulation_conditions.iterrows():
        meas_df_cond = measurement_df[
            measurement_df[SIMULATION_CONDITION_ID]
            == condition_id[SIMULATION_CONDITION_ID]
        ]
        sim_res = simulations[condition_id[SIMULATION_CONDITION_ID]]
        # in each row, replace the "measurement" with the simulation value
        for index, row in meas_df_cond.iterrows():
            timepoint = row[TIME]
            observable_id = row[OBSERVABLE_ID]
            time_index = np.where(sim_res[TIME] == timepoint)[0][0]
            sim_value = sim_res[observable_id][time_index]
            meas_df_cond.at[index, MEASUREMENT] = sim_value
        # rename measurement to simulation
        meas_df_cond = meas_df_cond.rename(columns={MEASUREMENT: SIMULATION})
        meas_dfs.append(meas_df_cond)
    sim_res_df = pd.concat(meas_dfs)
    return sim_res_df



"""
RoadRunner objective
====================
"""

__all__ = [
    "PetabImporterRR",
    "RoadRunnerCalculator",
    "ExpData",
    "SolverOptions",
]

from .petab_importer_roadrunner import PetabImporterRR
from .road_runner import RoadRunnerObjective
from .roadrunner_calculator import RoadRunnerCalculator
from .utils import ExpData, SolverOptions, simulation_to_measurement_df



"""Inner optimization problem in hierarchical optimization."""

import logging

import pandas as pd

from ...C import (
    MEASUREMENT_TYPE,
    PARAMETER_TYPE,
    SEMIQUANTITATIVE,
    InnerParameterType,
)
from ..base_problem import (
    AmiciInnerProblem,
    _get_timepoints_with_replicates,
    ix_matrices_from_arrays,
)
from .parameter import RelativeInnerParameter

try:
    import amici.sim.sundials as asd
    import petab.v1 as petab
    from petab.v1.C import (
        ESTIMATE,
        LOWER_BOUND,
        NOISE_PARAMETERS,
        OBSERVABLE_ID,
        OBSERVABLE_PARAMETERS,
        PARAMETER_ID,
        PARAMETER_SCALE,
        TIME,
        UPPER_BOUND,
    )
except ImportError:
    pass

logger = logging.getLogger(__name__)


class RelativeInnerProblem(AmiciInnerProblem):
    r"""Inner optimization problem for relative data with scaling/offset.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    edatas:
        AMICI ``ExpData``\s for each simulation condition.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def from_petab_amici(
        petab_problem: "petab.Problem",
        amici_model: "asd.Model",
        edatas: list["asd.ExpData"],
    ) -> "RelativeInnerProblem":
        """Create an InnerProblem from a PEtab problem and AMICI objects."""
        return inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas
        )

    def get_relative_observable_ids(self) -> list[str]:
        """Get IDs of all unique relative observables with scaling and/or offset."""
        return list(
            {
                observable_id
                for x in self.xs.values()
                if x.inner_parameter_type
                in [
                    InnerParameterType.SCALING,
                    InnerParameterType.OFFSET,
                ]
                for observable_id in x.observable_ids
            }
        )

    def get_observable_indices_for_xs(
        self, inner_parameter_type: str
    ) -> list[int]:
        """Get unique list of ``RelativeParameter.observable_indices`` values."""
        return list(
            {
                obs_idx
                for x in self.xs.values()
                if x.inner_parameter_type == inner_parameter_type
                for obs_idx in x.observable_indices
            }
        )

    def get_xs_for_obs_idx(self, obs_idx: int) -> list[RelativeInnerParameter]:
        r"""Get ``RelativeParameter``\s that belong to the observable with index `obs_idx`."""
        return [x for x in self.xs.values() if obs_idx in x.observable_indices]


def inner_problem_from_petab_problem(
    petab_problem: "petab.Problem",
    amici_model: "asd.Model",
    edatas: list["asd.ExpData"],
) -> RelativeInnerProblem:
    """
    Create inner problem from PEtab problem.

    Hierarchical optimization is a pypesto-specific PEtab extension.
    """

    # inner parameters
    inner_parameters = inner_parameters_from_parameter_df(
        petab_problem.parameter_df, petab_problem.measurement_df
    )

    x_ids = [x.inner_parameter_id for x in inner_parameters]

    # used indices for all measurement specific parameters
    ixs = ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, x_ids
    )

    # transform experimental data
    data = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, data)

    # assign matrices, observable indices and ids to inner parameters
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]
        par.observable_indices = [
            meas_indices[2] for meas_indices in ixs[par.inner_parameter_id]
        ]
        par.observable_ids = [
            amici_model.get_observable_ids()[obs_idx]
            for obs_idx in par.observable_indices
        ]

    par_group_types = {
        tuple(obs_pars.split(";")): (
            petab_problem.parameter_df.loc[obs_par, PARAMETER_TYPE]
            for obs_par in obs_pars.split(";")
        )
        for (obs_id, obs_pars), _ in petab_problem.measurement_df.groupby(
            [petab.OBSERVABLE_ID, petab.OBSERVABLE_PARAMETERS], dropna=True
        )
        if ";" in obs_pars  # prefilter for at least 2 observable parameters
    }

    coupled_pars = {
        group
        for group, types in par_group_types.items()
        if (
            (InnerParameterType.SCALING in types)
            and (InnerParameterType.OFFSET in types)
        )
    }

    # Check each group is of length 2
    for group in coupled_pars:
        if len(group) != 2:
            raise ValueError(
                f"Expected exactly 2 parameters in group {group}: a scaling "
                f"and an offset parameter."
            )

    id_to_par = {par.inner_parameter_id: par for par in inner_parameters}

    # assign coupling
    for par in inner_parameters:
        if par.inner_parameter_type not in [
            InnerParameterType.SCALING,
            InnerParameterType.OFFSET,
        ]:
            continue
        for group in coupled_pars:
            if par.inner_parameter_id in group:
                coupled_parameter_id = group[
                    group.index(par.inner_parameter_id) - 1
                ]
                par.coupled = id_to_par[coupled_parameter_id]
                break

    return RelativeInnerProblem(xs=inner_parameters, data=data, edatas=edatas)


def inner_parameters_from_parameter_df(
    par_df: pd.DataFrame,
    meas_df: pd.DataFrame,
) -> list[RelativeInnerParameter]:
    """
    Create list of inner free parameters from PEtab parameter table.

    Inner parameters are those that have a non-empty `parameterType` in the
    PEtab problem.
    """
    # create list of hierarchical parameters
    par_df = par_df.reset_index()

    for col in (PARAMETER_TYPE,):
        if col not in par_df:
            par_df[col] = None

    parameters = []

    for _, row in par_df.iterrows():
        if not row[ESTIMATE]:
            continue
        if petab.is_empty(row[PARAMETER_TYPE]):
            continue
        # If a sigma parameter belongs to a semiquantitative
        # observable, it is not a relative inner parameter.
        if row[PARAMETER_TYPE] == InnerParameterType.SIGMA:
            if MEASUREMENT_TYPE in meas_df.columns:
                par_id = row[PARAMETER_ID]
                corresponding_measurements = meas_df[
                    meas_df[NOISE_PARAMETERS] == par_id
                ]
                if any(
                    corresponding_measurements[MEASUREMENT_TYPE]
                    == SEMIQUANTITATIVE
                ):
                    continue

        parameters.append(
            RelativeInnerParameter(
                inner_parameter_id=row[PARAMETER_ID],
                inner_parameter_type=row[PARAMETER_TYPE],
                scale=row[PARAMETER_SCALE],
                lb=row[LOWER_BOUND],
                ub=row[UPPER_BOUND],
                observable_ids=None,
                observable_indices=None,
            )
        )

    return parameters


def ixs_for_measurement_specific_parameters(
    petab_problem: "petab.Problem",
    amici_model: "asd.Model",
    x_ids: list[str],
) -> dict[str, list[tuple[int, int, int]]]:
    """
    Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a list of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.get_observable_ids()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=petab_problem.measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                observable_overrides = petab.split_parameter_replacement_list(
                    measurement.get(OBSERVABLE_PARAMETERS, None)
                )
                noise_overrides = petab.split_parameter_replacement_list(
                    measurement.get(NOISE_PARAMETERS, None)
                )

                # try to insert if hierarchical parameter
                for override in observable_overrides + noise_overrides:
                    if override in x_ids:
                        ixs_for_par.setdefault(override, []).append(
                            (condition_ix, time_w_reps_ix, observable_ix)
                        )
    return ixs_for_par



import logging
from collections.abc import Callable
from typing import Literal

import numpy as np

from ..problem import Problem
from ..result import ProfilerResult
from .options import ProfileOptions

logger = logging.getLogger(__name__)

__all__ = ["next_guess", "fixed_step", "adaptive_step"]


def next_guess(
    x: np.ndarray,
    par_index: int,
    par_direction: Literal[1, -1],
    profile_options: ProfileOptions,
    update_type: Literal[
        "fixed_step",
        "adaptive_step_order_0",
        "adaptive_step_order_1",
        "adaptive_step_regression",
    ],
    current_profile: ProfilerResult,
    problem: Problem,
    global_opt: float,
    min_step_increase_factor: float = 1.0,
    max_step_reduce_factor: float = 1.0,
) -> np.ndarray:
    """
    Create the next initial guess for the optimizer.

    Used in order to compute the next profile point. Different proposal methods
    are available.

    Parameters
    ----------
    x:
        The current position of the profiler.
    par_index:
        The index of the parameter of the current profile.
    par_direction:
        The direction, in which the profiling is done (``1`` or ``-1``).
    profile_options:
        Various options applied to the profile optimization.
    update_type:
        Type of update for next profile point. Available options are:

        * ``fixed_step`` (see :func:`fixed_step`)
        * ``adaptive_step_order_0`` (see :func:`adaptive_step`).
        * ``adaptive_step_order_1`` (see :func:`adaptive_step`).
        * ``adaptive_step_regression`` (see :func:`adaptive_step`).
    current_profile:
        The profile which should be computed.
    problem:
        The problem to be solved.
    global_opt:
        Log-posterior value of the global optimum.
    min_step_increase_factor:
        Factor to increase the minimal step size bound. Used only in
        :func:`adaptive_step`.
    max_step_reduce_factor:
        Factor to reduce the maximal step size bound. Used only in
        :func:`adaptive_step`.

    Returns
    -------
    The next initial guess as base for the next profile point.
    """
    if update_type == "fixed_step":
        next_initial_guess = fixed_step(
            x, par_index, par_direction, profile_options, problem
        )
    elif update_type == "adaptive_step_order_0":
        order = 0
    elif update_type == "adaptive_step_order_1":
        order = 1
    elif update_type == "adaptive_step_regression":
        order = np.nan
    else:
        raise ValueError(
            f"Unsupported `update_type` {update_type} for `next_guess`."
        )
    if update_type != "fixed_step":
        next_initial_guess = adaptive_step(
            x,
            par_index,
            par_direction,
            profile_options,
            current_profile,
            problem,
            global_opt,
            order,
            min_step_increase_factor,
            max_step_reduce_factor,
        )

    logger.debug(
        f"Next guess for {problem.x_names[par_index]} in direction "
        f"{par_direction} is {next_initial_guess[par_index]:.4f}. Step size: "
        f"{next_initial_guess[par_index] - x[par_index]:.4f}."
    )

    return next_initial_guess


def fixed_step(
    x: np.ndarray,
    par_index: int,
    par_direction: Literal[1, -1],
    options: ProfileOptions,
    problem: Problem,
) -> np.ndarray:
    """Most simple method to create the next guess.

    Computes the next point based on the fixed step size given by
    :attr:`pypesto.profile.ProfileOptions.default_step_size`.

    Parameters
    ----------
    x:
       The current position of the profiler, size `dim_full`.
    par_index:
        The index of the parameter of the current profile.
    par_direction:
        The direction, in which the profiling is done (``1`` or ``-1``).
    options:
        Various options applied to the profile optimization.
    problem:
        The problem to be solved.

    Returns
    -------
    The updated parameter vector, of size `dim_full`.
    """
    delta_x = np.zeros(len(x))
    delta_x[par_index] = par_direction * options.default_step_size

    # check whether the next point is maybe outside the bounds
    # and correct it
    next_x_par = x[par_index] + delta_x[par_index]
    if par_direction == -1 and next_x_par < problem.lb_full[par_index]:
        delta_x[par_index] = problem.lb_full[par_index] - x[par_index]
    elif par_direction == 1 and next_x_par > problem.ub_full[par_index]:
        delta_x[par_index] = problem.ub_full[par_index] - x[par_index]

    return x + delta_x


def adaptive_step(
    x: np.ndarray,
    par_index: int,
    par_direction: Literal[1, -1],
    options: ProfileOptions,
    current_profile: ProfilerResult,
    problem: Problem,
    global_opt: float,
    order: int = 1,
    min_step_increase_factor: float = 1.0,
    max_step_reduce_factor: float = 1.0,
) -> np.ndarray:
    """Group of more complex methods for point proposal.

    Step size is automatically computed by a line search algorithm (hence:
    adaptive).

    Parameters
    ----------
    x:
        The current position of the profiler, size `dim_full`.
    par_index:
        The index of the parameter of the current profile.
    par_direction:
        The direction, in which the profiling is done (``1`` or ``-1``).
    options:
        Various options applied to the profile optimization.
    current_profile:
        The profile which should be computed.
    problem:
        The problem to be solved.
    global_opt:
        Log-posterior value of the global optimum.
    order:
        Specifies the precise algorithm for extrapolation.
        Available options are:

        * ``0``: just one parameter is updated
        * ``1``: the last two points are used to extrapolate all parameters
        * ``np.nan``: indicates that a more complex regression should be used
          as determined by :attr:`pypesto.profile.ProfileOptions.reg_order`.
    min_step_increase_factor:
        Factor to increase the minimal step size bound.
    max_step_reduce_factor:
        Factor to reduce the maximal step size bound.


    Returns
    -------
    The updated parameter vector, of size `dim_full`.
    """

    # restrict step proposal to minimum and maximum step size
    def clip_to_minmax(step_size_proposal):
        min_step_size = options.min_step_size * min_step_increase_factor
        max_step_size = options.max_step_size * max_step_reduce_factor
        return np.clip(step_size_proposal, min_step_size, max_step_size)

    # restrict step proposal to bounds
    def clip_to_bounds(step_proposal):
        return np.clip(step_proposal, problem.lb_full, problem.ub_full)

    problem.fix_parameters(par_index, x[par_index])

    # Get update directions and first step size guesses
    (
        step_size_guess,
        delta_x_dir,
        reg_par,
        delta_obj_value,
        last_delta_fval,
    ) = handle_profile_history(
        x,
        par_index,
        par_direction,
        global_opt,
        order,
        current_profile,
        problem,
        options,
    )

    # check whether we must make a minimum step anyway, since we're close to
    # the next bound
    min_delta_x = (
        x[par_index]
        + par_direction * options.min_step_size * min_step_increase_factor
    )

    if par_direction == -1 and (min_delta_x < problem.lb_full[par_index]):
        step_length = abs(problem.lb_full[par_index] - x[par_index])
        return clip_to_bounds(x + step_length * delta_x_dir)

    if par_direction == 1 and (min_delta_x > problem.ub_full[par_index]):
        step_length = abs(problem.ub_full[par_index] - x[par_index])
        return clip_to_bounds(x + step_length * delta_x_dir)

    # parameter extrapolation function
    n_profile_points = len(current_profile.fval_path)

    # Do we have enough points to do a regression?
    if np.isnan(order) and n_profile_points > 2:

        def par_extrapol(step_length):
            x_step = []
            # loop over parameters, extrapolate each one
            for i_par in range(problem.dim_full):
                if i_par == par_index:
                    # if we meet the profiling parameter, just increase,
                    # don't extrapolate
                    x_step.append(x[par_index] + step_length * par_direction)
                elif i_par in problem.x_fixed_indices:
                    # common fixed parameter: will be ignored anyway later
                    x_step.append(np.nan)
                else:
                    # extrapolate
                    cur_par_extrapol = np.poly1d(reg_par[i_par])
                    x_step.append(
                        cur_par_extrapol(
                            x[par_index] + step_length * par_direction
                        )
                    )
            # Define a trust region for the step size in all directions
            # to avoid overshooting
            x_step = np.clip(
                x_step, x - options.max_step_size, x + options.max_step_size
            )

            return clip_to_bounds(x_step)

    else:
        # if not, we do simple extrapolation
        def par_extrapol(step_length):
            # Define a trust region for the step size in all directions
            # to avoid overshooting
            step_in_x = np.clip(
                step_length * delta_x_dir,
                -options.max_step_size,
                options.max_step_size,
            )
            x_stepped = x + step_in_x
            return clip_to_bounds(x_stepped)

    # compute proposal
    next_x = par_extrapol(step_size_guess)

    # next start point has to be searched
    # compute the next objective value which we aim for
    high_next_obj_target = (
        -np.log(1.0 - options.delta_ratio_max)
        + options.adaptive_target_scaling_factor * abs(last_delta_fval)
        + current_profile.fval_path[-1]
    )
    low_next_obj_target = (
        +np.log(1.0 - options.delta_ratio_max)
        - options.adaptive_target_scaling_factor * abs(last_delta_fval)
        + current_profile.fval_path[-1]
    )

    # Clip both by 0.5 * delta_obj_value to avoid overshooting
    if delta_obj_value != 0:
        high_next_obj_target = min(
            high_next_obj_target,
            current_profile.fval_path[-1] + 0.5 * delta_obj_value,
        )
        low_next_obj_target = max(
            low_next_obj_target,
            current_profile.fval_path[-1] - 0.5 * delta_obj_value,
        )

    # compute objective at the guessed point
    problem.fix_parameters(par_index, next_x[par_index])
    next_obj = problem.objective(problem.get_reduced_vector(next_x))
    current_obj = current_profile.fval_path[-1]

    # iterate until good step size is found
    return do_line_search(
        next_x,
        step_size_guess,
        par_extrapol,
        next_obj,
        current_obj,
        high_next_obj_target,
        low_next_obj_target,
        clip_to_minmax,
        clip_to_bounds,
        par_index,
        problem,
        options,
        min_step_increase_factor,
        max_step_reduce_factor,
    )


def handle_profile_history(
    x: np.ndarray,
    par_index: int,
    par_direction: Literal[1, -1],
    global_opt: float,
    order: int,
    current_profile: ProfilerResult,
    problem: Problem,
    options: ProfileOptions,
) -> tuple[float, np.array, list[float], float]:
    """Compute the very first step direction update guesses.

    Check whether enough steps have been taken for applying regression,
    computes regression or simple extrapolation.

    Returns
    -------
    step_size_guess:
        Guess for the step size.
    delta_x_dir:
        Parameter update direction.
    reg_par:
        The regression polynomial for profile extrapolation.
    delta_obj_value:
        The difference of the objective function value between the last point and `global_opt`.
    last_delta_fval:
        The difference of the objective function value between the last two points.
    """
    n_profile_points = len(current_profile.fval_path)

    # set the update direction
    delta_x_dir = np.zeros(len(x))
    delta_x_dir[par_index] = par_direction
    reg_par = None

    # Is this the first step along this profile? If so, try a simple step
    # Do the same if the last two points are too close to avoid division by small numbers
    if n_profile_points == 1 or np.isclose(
        current_profile.x_path[par_index, -1],
        current_profile.x_path[par_index, -2],
    ):
        # try to use the default step size
        step_size_guess = options.default_step_size
        delta_obj_value = 0.0
        last_delta_fval = 0.0

    else:
        # try to reuse the previous step size
        last_delta_x_par_index = np.abs(
            current_profile.x_path[par_index, -1]
            - current_profile.x_path[par_index, -2]
        )
        # Bound the step size by default values
        step_size_guess = min(
            last_delta_x_par_index, options.default_step_size
        )
        # Step size cannot be smaller than the minimum step size
        step_size_guess = max(step_size_guess, options.min_step_size)

        delta_obj_value = current_profile.fval_path[-1] - global_opt
        last_delta_fval = (
            current_profile.fval_path[-1] - current_profile.fval_path[-2]
        )

        if order == 1 or (np.isnan(order) and n_profile_points < 3):
            # set the update direction (extrapolate with order 1)
            last_delta_x = (
                current_profile.x_path[:, -1] - current_profile.x_path[:, -2]
            )
            delta_x_dir = last_delta_x / last_delta_x_par_index
        elif np.isnan(order):
            # compute the regression polynomial for parameter extrapolation
            reg_par = get_reg_polynomial(
                par_index, current_profile, problem, options
            )

    return (
        step_size_guess,
        delta_x_dir,
        reg_par,
        delta_obj_value,
        last_delta_fval,
    )


def get_reg_polynomial(
    par_index: int,
    current_profile: ProfilerResult,
    problem: Problem,
    options: ProfileOptions,
) -> list[float]:
    """Compute the regression polynomial.

    Used to step proposal extrapolation from the last profile points.
    """
    # determine interpolation order
    n_profile_points = len(current_profile.fval_path)
    reg_max_order = np.floor(n_profile_points / 2)
    reg_order = min(reg_max_order, options.reg_order)
    reg_points = min(n_profile_points, options.reg_points)

    # set up matrix of regression parameters
    reg_par = []
    for i_par in range(problem.dim_full):
        if i_par in problem.x_fixed_indices:
            # if we meet the current profiling parameter or a fixed parameter,
            # there is nothing to do, so pass a np.nan
            reg_par.append(np.nan)
        else:
            # Do polynomial interpolation of profile path
            # Determine rank of polynomial interpolation
            regression_tmp = np.polyfit(
                current_profile.x_path[par_index, -reg_points:],
                current_profile.x_path[i_par, -reg_points:],
                reg_order,
                full=True,
            )

            # Decrease rank if interpolation problem is ill-conditioned
            if regression_tmp[2] < reg_order:
                reg_order = regression_tmp[2]
                regression_tmp = np.polyfit(
                    current_profile.x_path[par_index, -reg_points:],
                    current_profile.x_path[i_par, -reg_points:],
                    int(reg_order),
                    full=True,
                )

            # add to regression parameters
            reg_par.append(regression_tmp[0])

    return reg_par


def do_line_search(
    next_x: np.ndarray,
    step_size_guess: float,
    par_extrapol: Callable,
    next_obj: float,
    current_obj: float,
    high_next_obj_target: float,
    low_next_obj_target: float,
    clip_to_minmax: Callable,
    clip_to_bounds: Callable,
    par_index: int,
    problem: Problem,
    options: ProfileOptions,
    min_step_increase_factor: float,
    max_step_reduce_factor: float,
) -> np.ndarray:
    """Perform the line search.

    Based on the objective function we want to reach, based on the current
    position in parameter space and on the first guess for the proposal.

    Parameters
    ----------
    next_x:
        Starting parameters for the line search.
    step_size_guess:
        First guess for the step size.
    par_extrapol:
        Parameter extrapolation function.
    next_obj:
        Objective function value at `next_x`.
    next_obj_target:
        Objective function value we want to reach.
    clip_to_minmax:
        Function to clip the step size to minimum and maximum step size.
    clip_to_bounds:
        Function to clip the parameters to the bounds.
    par_index:
        Index of the parameter we are profiling.
    problem:
        The parameter estimation problem.
    options:
        Profile likelihood options.
    min_step_increase_factor:
        Factor to increase the minimal step size bound.
    max_step_reduce_factor:
        Factor to reduce the maximal step size bound.

    Returns
    -------
    Parameter vector that is expected to yield the objective function value
    closest to `next_obj_target`.
    """
    decreasing_to_low_target = False
    decreasing_to_high_target = False

    # Determine the direction of the step
    if next_obj > low_next_obj_target and next_obj < high_next_obj_target:
        direction = "increase"
    elif next_obj <= low_next_obj_target:
        direction = "decrease"
        decreasing_to_low_target = True
    elif next_obj >= high_next_obj_target:
        direction = "decrease"
        decreasing_to_high_target = True

    if direction == "increase":
        adapt_factor = options.step_size_factor
    else:
        adapt_factor = 1 / options.step_size_factor

    # Loop until correct step size was found
    while True:
        # Adapt step size of guess
        last_x = next_x
        step_size_guess = clip_to_minmax(step_size_guess * adapt_factor)
        next_x = clip_to_bounds(par_extrapol(step_size_guess))

        # Check if we hit the bounds
        if (
            direction == "decrease"
            and step_size_guess
            == options.min_step_size * min_step_increase_factor
        ):
            return next_x
        if (
            direction == "increase"
            and step_size_guess
            == options.max_step_size * max_step_reduce_factor
        ):
            return next_x

        # compute new objective value
        problem.fix_parameters(par_index, next_x[par_index])
        last_obj = next_obj
        next_obj = problem.objective(problem.get_reduced_vector(next_x))

        # check for root crossing and compute correct step size in case
        if (direction == "increase" and next_obj > high_next_obj_target) or (
            direction == "decrease"
            and next_obj < high_next_obj_target
            and decreasing_to_high_target
        ):
            return next_x_interpolate(
                next_obj, last_obj, next_x, last_x, high_next_obj_target
            )

        if (direction == "increase" and next_obj < low_next_obj_target) or (
            direction == "decrease"
            and next_obj > low_next_obj_target
            and decreasing_to_low_target
        ):
            return next_x_interpolate(
                next_obj, last_obj, next_x, last_x, low_next_obj_target
            )


def next_x_interpolate(
    next_obj: float,
    last_obj: float,
    next_x: np.ndarray,
    last_x: np.ndarray,
    next_obj_target: float,
) -> np.ndarray:
    """Interpolate between the last two steps."""
    delta_obj = np.abs(next_obj - last_obj)
    add_x = np.abs(last_obj - next_obj_target) * (next_x - last_x) / delta_obj

    # fix final guess and return
    return last_x + add_x



"""
Relative data integration
=========================

Contains an implementation of the hierarchical inner subproblem for relative data.
In this inner problem, the scaling factors, offsets, and noise standard deviations
are optimized, conditional on the outer dynamical parameters. The inner problem
can be solved analytically.

An example of parameter estimation with relative data
can be found in pypesto/doc/examples/relative_data.ipynb.

The implementation in this package is based on:

* Loos et al. 2018 (https://doi.org/10.1093/bioinformatics/bty514),
  who give an analytic solution for the inner problem for scaling factors and
  noise standard deviations, for Gaussian and Laplace noise, using forward
  sensitivity analysis (FSA).
* Schmiester et al. 2020 (https://doi.org/10.1093/bioinformatics/btz581),
  who give an analytic solution for the inner problem for scaling factors,
  offsets and noise standard deviations, for Gaussian and Laplace noise,
  using adjoint sensitivity analysis (ASA). ASA allows to calculate gradients
  substantially more efficiently in high dimension.
"""

from .calculator import RelativeAmiciCalculator
from .problem import RelativeInnerProblem
from .solver import AnalyticalInnerSolver, NumericalInnerSolver
